from __future__ import annotations

import logging
import threading
import time
import traceback
from concurrent.futures.thread import ThreadPoolExecutor
from pathlib import Path
from typing import Any, Callable, Optional

from chik_rs import G1Element
from chikpos import DiskProver, decompressor_context_queue

from chik.consensus.pos_quality import UI_ACTUAL_SPACE_CONSTANT_FACTOR, _expected_plot_size
from chik.plotting.cache import Cache, CacheEntry
from chik.plotting.util import (
    HarvestingMode,
    PlotInfo,
    PlotRefreshEvents,
    PlotRefreshResult,
    PlotsRefreshParameter,
    get_plot_filenames,
)
from chik.util.batches import to_batches

log = logging.getLogger(__name__)


class PlotManager:
    plots: dict[Path, PlotInfo]
    plot_filename_paths: dict[str, tuple[str, set[str]]]
    plot_filename_paths_lock: threading.Lock
    failed_to_open_filenames: dict[Path, int]
    no_key_filenames: set[Path]
    farmer_public_keys: list[G1Element]
    pool_public_keys: list[G1Element]
    cache: Cache
    match_str: Optional[str]
    open_no_key_filenames: bool
    last_refresh_time: float
    refresh_parameter: PlotsRefreshParameter
    log: Any
    _lock: threading.Lock
    _refresh_thread: Optional[threading.Thread]
    _refreshing_enabled: bool
    _refresh_callback: Callable
    _initial: bool
    max_compression_level_allowed: int
    context_count: int

    def __init__(
        self,
        root_path: Path,
        refresh_callback: Callable,
        match_str: Optional[str] = None,
        open_no_key_filenames: bool = False,
        refresh_parameter: PlotsRefreshParameter = PlotsRefreshParameter(),
    ):
        self.root_path = root_path
        self.plots = {}
        self.plot_filename_paths = {}
        self.plot_filename_paths_lock = threading.Lock()
        self.failed_to_open_filenames = {}
        self.no_key_filenames = set()
        self.farmer_public_keys = []
        self.pool_public_keys = []
        # Since `compression_level` property was added to Cache structure,
        # previous cache file formats needs to be reset
        # When user downgrades harvester, it looks 'plot_manager.dat` while
        # latest harvester reads/writes 'plot_manager_v2.dat`
        self.cache = Cache(self.root_path.resolve() / "cache" / "plot_manager_v2.dat")
        self.match_str = match_str
        self.open_no_key_filenames = open_no_key_filenames
        self.last_refresh_time = 0
        self.refresh_parameter = refresh_parameter
        self.log = logging.getLogger(__name__)
        self._lock = threading.Lock()
        self._refresh_thread = None
        self._refreshing_enabled = False
        self._refresh_callback = refresh_callback
        self._initial = True
        self.max_compression_level_allowed = 0
        self.context_count = 0

    def __enter__(self):
        self._lock.acquire()

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self._lock.release()

    def configure_decompressor(
        self,
        context_count: int,
        thread_count: int,
        disable_cpu_affinity: bool,
        max_compression_level_allowed: int,
        use_gpu_harvesting: bool,
        gpu_index: int,
        enforce_gpu_index: bool,
        decompressor_timeout: int,
    ) -> HarvestingMode:
        if max_compression_level_allowed > 7:
            log.error(
                "Currently only compression levels up to 7 are allowed, "
                f"while {max_compression_level_allowed} was specified."
                "Setting max_compression_level_allowed to 7."
            )
            max_compression_level_allowed = 7
        is_using_gpu = decompressor_context_queue.init(
            context_count,
            thread_count,
            disable_cpu_affinity,
            max_compression_level_allowed,
            use_gpu_harvesting,
            gpu_index,
            enforce_gpu_index,
            decompressor_timeout,
        )
        if not is_using_gpu and use_gpu_harvesting:
            log.error(
                "GPU harvesting failed initialization. "
                f"Falling back to CPU harvesting: {context_count} decompressors count, {thread_count} threads."
            )
        self.max_compression_level_allowed = max_compression_level_allowed
        self.context_count = context_count
        return HarvestingMode.GPU if is_using_gpu else HarvestingMode.CPU

    def reset(self) -> None:
        with self:
            self.last_refresh_time = time.time()
            self.plots.clear()
            self.plot_filename_paths.clear()
            self.failed_to_open_filenames.clear()
            self.no_key_filenames.clear()
            self._initial = True

    def set_refresh_callback(self, callback: Callable):
        self._refresh_callback = callback

    def set_public_keys(self, farmer_public_keys: list[G1Element], pool_public_keys: list[G1Element]):
        self.farmer_public_keys = farmer_public_keys
        self.pool_public_keys = pool_public_keys

    def initial_refresh(self) -> bool:
        return self._initial

    def public_keys_available(self) -> bool:
        return len(self.farmer_public_keys) > 0 and len(self.pool_public_keys) > 0

    def plot_count(self) -> int:
        with self:
            return len(self.plots)

    def get_duplicates(self) -> list[Path]:
        result = []
        for plot_filename, paths_entry in self.plot_filename_paths.items():
            _, duplicated_paths = paths_entry
            for path in duplicated_paths:
                result.append(Path(path) / plot_filename)
        return result

    def needs_refresh(self) -> bool:
        return time.time() - self.last_refresh_time > float(self.refresh_parameter.interval_seconds)

    def start_refreshing(self, sleep_interval_ms: int = 1000):
        self._refreshing_enabled = True
        if self._refresh_thread is None or not self._refresh_thread.is_alive():
            self.cache.load()
            self._refresh_thread = threading.Thread(target=self._refresh_task, args=(sleep_interval_ms,))
            self._refresh_thread.start()

    def stop_refreshing(self) -> None:
        self._refreshing_enabled = False
        if self._refresh_thread is not None and self._refresh_thread.is_alive():
            self._refresh_thread.join()
            self._refresh_thread = None

    def trigger_refresh(self) -> None:
        log.debug("trigger_refresh")
        self.last_refresh_time = 0

    def _refresh_task(self, sleep_interval_ms: int):
        while self._refreshing_enabled:
            try:
                while not self.needs_refresh() and self._refreshing_enabled:
                    time.sleep(sleep_interval_ms / 1000.0)

                if not self._refreshing_enabled:
                    return

                plot_filenames: dict[Path, list[Path]] = get_plot_filenames(self.root_path)
                plot_directories: set[Path] = set(plot_filenames.keys())
                plot_paths: set[Path] = set()
                for paths in plot_filenames.values():
                    plot_paths.update(paths)

                total_result: PlotRefreshResult = PlotRefreshResult()
                total_size = len(plot_paths)

                self._refresh_callback(PlotRefreshEvents.started, PlotRefreshResult(remaining=total_size))

                # First drop all plots we have in plot_filename_paths but not longer in the filesystem or set in config
                for path in list(self.failed_to_open_filenames.keys()):
                    if path not in plot_paths:
                        del self.failed_to_open_filenames[path]

                for path in self.no_key_filenames.copy():
                    if path not in plot_paths:
                        self.no_key_filenames.remove(path)

                filenames_to_remove: list[str] = []
                for plot_filename, paths_entry in self.plot_filename_paths.items():
                    loaded_path, duplicated_paths = paths_entry
                    loaded_plot = Path(loaded_path) / Path(plot_filename)
                    if loaded_plot not in plot_paths:
                        filenames_to_remove.append(plot_filename)
                        with self:
                            if loaded_plot in self.plots:
                                del self.plots[loaded_plot]
                        total_result.removed.append(loaded_plot)
                        # No need to check the duplicates here since we drop the whole entry
                        continue

                    paths_to_remove: list[str] = []
                    for path_str in duplicated_paths:
                        loaded_plot = Path(path_str) / Path(plot_filename)
                        if loaded_plot not in plot_paths:
                            paths_to_remove.append(path_str)
                    for path_str in paths_to_remove:
                        duplicated_paths.remove(path_str)

                for filename in filenames_to_remove:
                    del self.plot_filename_paths[filename]

                for batch in to_batches(sorted(list(plot_paths)), self.refresh_parameter.batch_size):
                    batch_result: PlotRefreshResult = self.refresh_batch(batch.entries, plot_directories)
                    if not self._refreshing_enabled:
                        self.log.debug("refresh_plots: Aborted")
                        break
                    # Set the remaining files since `refresh_batch()` doesn't know them but we want to report it
                    batch_result.remaining = batch.remaining
                    total_result.loaded += batch_result.loaded
                    total_result.processed += batch_result.processed
                    total_result.duration += batch_result.duration

                    self._refresh_callback(PlotRefreshEvents.batch_processed, batch_result)
                    if batch.remaining == 0:
                        break
                    batch_sleep = self.refresh_parameter.batch_sleep_milliseconds
                    self.log.debug(f"refresh_plots: Sleep {batch_sleep} milliseconds")
                    time.sleep(float(batch_sleep) / 1000.0)

                if self._refreshing_enabled:
                    self._refresh_callback(PlotRefreshEvents.done, total_result)

                # Reset the initial refresh indication
                self._initial = False

                # Cleanup unused cache
                self.log.debug(f"_refresh_task: cached entries before cleanup: {len(self.cache)}")
                remove_paths: list[Path] = []
                for path, cache_entry in self.cache.items():
                    if cache_entry.expired(Cache.expiry_seconds) and path not in self.plots:
                        remove_paths.append(path)
                    elif path in self.plots:
                        cache_entry.bump_last_use()
                self.cache.remove(remove_paths)
                self.log.debug(f"_refresh_task: cached entries removed: {len(remove_paths)}")

                if self.cache.changed():
                    self.cache.save()

                self.last_refresh_time = time.time()

                self.log.debug(
                    f"_refresh_task: total_result.loaded {len(total_result.loaded)}, "
                    f"total_result.removed {len(total_result.removed)}, "
                    f"total_duration {total_result.duration:.2f} seconds"
                )
            except Exception as e:
                log.error(f"_refresh_callback raised: {e} with the traceback: {traceback.format_exc()}")
                self.reset()

    def refresh_batch(self, plot_paths: list[Path], plot_directories: set[Path]) -> PlotRefreshResult:
        start_time: float = time.time()
        result: PlotRefreshResult = PlotRefreshResult(processed=len(plot_paths))
        counter_lock = threading.Lock()

        log.debug(f"refresh_batch: {len(plot_paths)} files in directories {plot_directories}")

        if self.match_str is not None:
            log.info(f'Only loading plots that contain "{self.match_str}" in the file or directory name')

        def process_file(file_path: Path) -> Optional[PlotInfo]:
            if not self._refreshing_enabled:
                return None
            filename_str = str(file_path)
            if self.match_str is not None and self.match_str not in filename_str:
                return None
            if (
                file_path in self.failed_to_open_filenames
                and (time.time() - self.failed_to_open_filenames[file_path])
                < self.refresh_parameter.retry_invalid_seconds
            ):
                # Try once every `refresh_parameter.retry_invalid_seconds` seconds to open the file
                return None

            if file_path in self.plots:
                return self.plots[file_path]

            entry: Optional[tuple[str, set[str]]] = self.plot_filename_paths.get(file_path.name)
            if entry is not None:
                _loaded_parent, duplicates = entry
                if str(file_path.parent) in duplicates:
                    log.debug(f"Skip duplicated plot {file_path!s}")
                    return None
            try:
                if not file_path.exists():
                    return None

                stat_info = file_path.stat()

                cache_entry = self.cache.get(file_path)
                cache_hit = cache_entry is not None
                if not cache_hit:
                    prover = DiskProver(str(file_path))

                    log.debug(f"process_file {file_path!s}")

                    expected_size = _expected_plot_size(prover.get_size()) * UI_ACTUAL_SPACE_CONSTANT_FACTOR

                    # TODO: consider checking if the file was just written to (which would mean that the file is still
                    # being copied). A segfault might happen in this edge case.

                    level = prover.get_compression_level()
                    if level == 0:
                        if prover.get_size() >= 30 and stat_info.st_size < 0.98 * expected_size:
                            log.warning(
                                f"Not farming plot {file_path}. "
                                f"Size is {stat_info.st_size / (1024 ** 3)} GiB, "
                                f"but expected at least: {expected_size / (1024 ** 3)} GiB. "
                                "We assume the file is being copied."
                            )
                            return None

                    cache_entry = CacheEntry.from_disk_prover(prover)
                    self.cache.update(file_path, cache_entry)

                assert cache_entry is not None

                level = cache_entry.prover.get_compression_level()
                if level > self.max_compression_level_allowed:
                    log.warning(
                        f"Not farming plot {file_path}. Plot compression level: {level}, "
                        f"max compression level allowed: {self.max_compression_level_allowed}."
                    )
                    return None

                if level > 0 and self.context_count == 0:
                    log.warning(
                        f"Not farming compressed plot {file_path}. Plot compression level: {level}, "
                        f"because parallel_decompressor_count is set to 0 in config.yaml. Use a non-zero value"
                        " to start harvesting compressed plots."
                    )
                    return None

                # Only use plots that correct keys associated with them
                if cache_entry.farmer_public_key not in self.farmer_public_keys:
                    log.warning(f"Plot {file_path} has a farmer public key that is not in the farmer's pk list.")
                    self.no_key_filenames.add(file_path)
                    if not self.open_no_key_filenames:
                        return None

                if cache_entry.pool_public_key is not None and cache_entry.pool_public_key not in self.pool_public_keys:
                    log.warning(f"Plot {file_path} has a pool public key that is not in the farmer's pool pk list.")
                    self.no_key_filenames.add(file_path)
                    if not self.open_no_key_filenames:
                        return None

                # If a plot is in `no_key_filenames` the keys were missing in earlier refresh cycles. We can remove
                # the current plot from that list if its in there since we passed the key checks above.
                if file_path in self.no_key_filenames:
                    self.no_key_filenames.remove(file_path)

                with self.plot_filename_paths_lock:
                    paths: Optional[tuple[str, set[str]]] = self.plot_filename_paths.get(file_path.name)
                    if paths is None:
                        paths = (str(Path(cache_entry.prover.get_filename()).parent), set())
                        self.plot_filename_paths[file_path.name] = paths
                    else:
                        paths[1].add(str(Path(cache_entry.prover.get_filename()).parent))
                        log.warning(f"Have multiple copies of the plot {file_path.name} in {[paths[0], *paths[1]]}.")
                        return None

                new_plot_info: PlotInfo = PlotInfo(
                    cache_entry.prover,
                    cache_entry.pool_public_key,
                    cache_entry.pool_contract_puzzle_hash,
                    cache_entry.plot_public_key,
                    stat_info.st_size,
                    stat_info.st_mtime,
                )

                cache_entry.bump_last_use()

                with counter_lock:
                    result.loaded.append(new_plot_info)

                if file_path in self.failed_to_open_filenames:
                    del self.failed_to_open_filenames[file_path]

            except Exception as e:
                tb = traceback.format_exc()
                log.error(f"Failed to open file {file_path}. {e} {tb}")
                self.failed_to_open_filenames[file_path] = int(time.time())
                return None
            log.debug(f"Found plot {file_path} of size {new_plot_info.prover.get_size()}, cache_hit: {cache_hit}")

            return new_plot_info

        with self, ThreadPoolExecutor() as executor:
            plots_refreshed: dict[Path, PlotInfo] = {}
            for new_plot in executor.map(process_file, plot_paths):
                if new_plot is not None:
                    plots_refreshed[Path(new_plot.prover.get_filename())] = new_plot
            self.plots.update(plots_refreshed)

        result.duration = time.time() - start_time

        self.log.debug(
            f"refresh_batch: loaded {len(result.loaded)}, "
            f"removed {len(result.removed)}, processed {result.processed}, "
            f"remaining {result.remaining}, batch_size {self.refresh_parameter.batch_size}, "
            f"duration: {result.duration:.2f} seconds"
        )
        return result



from __future__ import annotations

import argparse
import dataclasses
import json
import logging
import types
from pathlib import Path
from typing import Any

import testconfig

root_path = Path(__file__).parent.absolute()
project_root_path = root_path.parent.parent


def skip(path: Path) -> bool:
    return any(part.startswith(("_", ".")) and part != "_tests" for part in path.parts)


def subdirs(per: str) -> list[Path]:
    if per == "directory":
        glob_pattern = "**/"
    elif per == "file":
        glob_pattern = "**/test_*.py"
    else:
        raise Exception(f"Unrecognized per: {per!r}")

    paths = [path for path in root_path.rglob(glob_pattern) if not skip(path=path)]

    if per == "directory":
        filtered_paths = []
        for path in paths:
            relative_path = path.relative_to(root_path)
            logging.info(f"Considering: {relative_path}")
            if len([f for f in path.glob("test_*.py")]) == 0:
                logging.info(f"Skipping {relative_path}: no tests collected")
                continue

            filtered_paths.append(path)

        paths = filtered_paths

    return sorted(paths)


def module_dict(module: types.ModuleType) -> dict[str, Any]:
    return {k: v for k, v in module.__dict__.items() if not k.startswith("_") and k != "annotations"}


def dir_config(dir: Path) -> dict[str, Any]:
    import importlib

    module_name = ".".join([*dir.relative_to(root_path).parts, "config"])
    try:
        return module_dict(importlib.import_module(module_name))
    except ModuleNotFoundError:
        return {}


@dataclasses.dataclass
class SpecifiedDefaultsError(Exception):
    overlap: dict[str, Any]

    def __post_init__(self) -> None:
        super().__init__()


# Overwrite with directory specific values
def update_config(parent: dict[str, Any], child: dict[str, Any]) -> dict[str, Any]:
    if child is None:
        return parent
    conf = child

    # avoid manual configuration set to default values
    common_keys = set(parent.keys()).intersection(child.keys())
    specified_defaulted_values = {k: parent[k] for k in common_keys if parent[k] == child[k]}
    if len(specified_defaulted_values) > 0:
        raise SpecifiedDefaultsError(overlap=specified_defaulted_values)

    for k, v in parent.items():
        if k not in child:
            conf[k] = v

    return conf


# args
arg_parser = argparse.ArgumentParser(description="Generate GitHub test matrix configuration")
arg_parser.add_argument("--per", type=str, choices=["directory", "file"], required=True)
arg_parser.add_argument("--verbose", "-v", action="store_true")
arg_parser.add_argument("--only", action="append", default=[])
arg_parser.add_argument("--duplicates", type=int, default=1)
arg_parser.add_argument("--timeout-multiplier", type=float, default=1)
args = arg_parser.parse_args()

if args.verbose:
    logging.basicConfig(format="%(asctime)s:%(message)s", level=logging.DEBUG)

# main
if len(args.only) == 0:
    test_paths = subdirs(per=args.per)
else:
    test_paths = [root_path.joinpath(path) for path in args.only]

test_paths_with_index = [(path, index + 1) for path in test_paths for index in range(args.duplicates)]

configuration = []

specified_defaults: dict[Path, dict[str, Any]] = {}
pytest_monitor_enabling_paths: list[Path] = []

for path, index in test_paths_with_index:
    if path.is_dir():
        test_files = sorted(path.glob("test_*.py"))
        paths_for_cli_list = [file.relative_to(project_root_path) for file in test_files]
        config_path = path
    else:
        paths_for_cli_list = [path.relative_to(project_root_path)]
        config_path = path.parent

    def mung_path(path: Path) -> str:
        # TODO: shell escaping, but that's per platform...
        return ".".join(path.with_suffix("").parts)

    paths_for_cli = " ".join(mung_path(path) for path in paths_for_cli_list)
    paths_for_cli = f"--pyargs {paths_for_cli}"

    try:
        conf = update_config(module_dict(testconfig), dir_config(config_path))
    except SpecifiedDefaultsError as e:
        specified_defaults[root_path.joinpath(config_path, "config.py")] = e.overlap
        continue

    # TODO: design a configurable system for this
    process_count = {
        "macos": {False: 0, True: 4}.get(conf["parallel"], conf["parallel"]),
        "ubuntu": {False: 0, True: 6}.get(conf["parallel"], conf["parallel"]),
        "windows": {False: 0, True: 4}.get(conf["parallel"], conf["parallel"]),
    }
    pytest_parallel_args = {os: f" -n {count}" for os, count in process_count.items()}

    enable_pytest_monitor = conf["check_resource_usage"]

    if enable_pytest_monitor:
        # NOTE: do not use until the hangs are fixed
        #       https://github.com/CFMTech/pytest-monitor/issues/53
        #       https://github.com/pythonprofilers/memory_profiler/issues/342

        pytest_monitor_enabling_paths.append(path)

    max_index_characters = len(str(args.duplicates))
    index_string = f"{index:0{max_index_characters}d}"
    module_import_path = ".".join(path.relative_to(root_path).with_suffix("").parts)
    if args.duplicates == 1:
        name = module_import_path
        file_name_index = ""
    else:
        name = f"{module_import_path} #{index_string}"
        file_name_index = f"_{index_string}"

    for_matrix = {
        "check_resource_usage": conf["check_resource_usage"],
        "enable_pytest_monitor": "-p monitor" if enable_pytest_monitor else "",
        "job_timeout": round(conf["job_timeout"] * args.timeout_multiplier),
        "pytest_parallel_args": pytest_parallel_args,
        "checkout_blocks_and_plots": conf["checkout_blocks_and_plots"],
        "install_timelord": conf["install_timelord"],
        "test_files": paths_for_cli,
        "name": name,
        "legacy_keyring_required": conf.get("legacy_keyring_required", False),
        "index": index,
        "index_string": index_string,
        "module_import_path": module_import_path,
        "file_name_index": file_name_index,
    }
    for_matrix = dict(sorted(for_matrix.items()))
    configuration.append(for_matrix)

messages: list[str] = []

if len(specified_defaults) > 0:
    message = f"Found {len(specified_defaults)} directories with specified defaults"
    messages.append(message)
    logging.error(f"{message}:")
    for path, overlap in sorted(specified_defaults.items()):
        logging.error(f" {path} : {overlap}")

if len(pytest_monitor_enabling_paths) > 0:
    message = f"Found {len(pytest_monitor_enabling_paths)} directories with pytest-monitor enabled"
    messages.append(message)
    logging.error(f"{message}:")
    for path in sorted(pytest_monitor_enabling_paths):
        logging.error(f" {path}")

if len(messages) > 0:
    raise Exception("\n".join(messages))

configuration_json = json.dumps(configuration)

for line in json.dumps(configuration, indent=2).splitlines():
    logging.info(line)

print(f"{configuration_json}")



from __future__ import annotations

import asyncio
import logging
import time
from collections.abc import Awaitable, Collection
from concurrent.futures import Executor, ThreadPoolExecutor
from dataclasses import dataclass
from typing import Callable, Optional, TypeVar

from chik_rs import ELIGIBLE_FOR_DEDUP, ELIGIBLE_FOR_FF, BLSCache, supports_fast_forward, validate_klvm_and_signature
from chikbip158 import PyBIP158

from chik.consensus.block_record import BlockRecordProtocol
from chik.consensus.constants import ConsensusConstants
from chik.consensus.cost_calculator import NPCResult
from chik.full_node.bitcoin_fee_estimator import create_bitcoin_fee_estimator
from chik.full_node.fee_estimation import FeeBlockInfo, MempoolInfo, MempoolItemInfo
from chik.full_node.fee_estimator_interface import FeeEstimatorInterface
from chik.full_node.mempool import MEMPOOL_ITEM_FEE_LIMIT, Mempool, MempoolRemoveInfo, MempoolRemoveReason
from chik.full_node.mempool_check_conditions import mempool_check_time_locks
from chik.full_node.pending_tx_cache import ConflictTxCache, PendingTxCache
from chik.types.blockchain_format.coin import Coin
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.klvm_cost import KLVMCost
from chik.types.coin_record import CoinRecord
from chik.types.eligible_coin_spends import EligibilityAndAdditions, UnspentLineageInfo
from chik.types.fee_rate import FeeRate
from chik.types.mempool_inclusion_status import MempoolInclusionStatus
from chik.types.mempool_item import BundleCoinSpend, MempoolItem
from chik.types.spend_bundle import SpendBundle
from chik.types.spend_bundle_conditions import SpendBundleConditions
from chik.util.db_wrapper import SQLITE_INT_MAX
from chik.util.errors import Err, ValidationError
from chik.util.inline_executor import InlineExecutor
from chik.util.ints import uint32, uint64

log = logging.getLogger(__name__)

# mempool items replacing existing ones must increase the total fee at least by
# this amount. 0.00001 XCK
MEMPOOL_MIN_FEE_INCREASE = uint64(10000000)


@dataclass
class TimelockConditions:
    assert_height: uint32 = uint32(0)
    assert_seconds: uint64 = uint64(0)
    assert_before_height: Optional[uint32] = None
    assert_before_seconds: Optional[uint64] = None


def compute_assert_height(
    removal_coin_records: dict[bytes32, CoinRecord],
    conds: SpendBundleConditions,
) -> TimelockConditions:
    """
    Computes the most restrictive height- and seconds assertion in the spend bundle.
    Relative heights and times are resolved using the confirmed heights and
    timestamps from the coin records.
    """

    ret = TimelockConditions()
    ret.assert_height = uint32(conds.height_absolute)
    ret.assert_seconds = uint64(conds.seconds_absolute)
    ret.assert_before_height = (
        uint32(conds.before_height_absolute) if conds.before_height_absolute is not None else None
    )
    ret.assert_before_seconds = (
        uint64(conds.before_seconds_absolute) if conds.before_seconds_absolute is not None else None
    )

    for spend in conds.spends:
        if spend.height_relative is not None:
            h = uint32(removal_coin_records[bytes32(spend.coin_id)].confirmed_block_index + spend.height_relative)
            ret.assert_height = max(ret.assert_height, h)

        if spend.seconds_relative is not None:
            s = uint64(removal_coin_records[bytes32(spend.coin_id)].timestamp + spend.seconds_relative)
            ret.assert_seconds = max(ret.assert_seconds, s)

        if spend.before_height_relative is not None:
            h = uint32(
                removal_coin_records[bytes32(spend.coin_id)].confirmed_block_index + spend.before_height_relative
            )
            if ret.assert_before_height is not None:
                ret.assert_before_height = min(ret.assert_before_height, h)
            else:
                ret.assert_before_height = h

        if spend.before_seconds_relative is not None:
            s = uint64(removal_coin_records[bytes32(spend.coin_id)].timestamp + spend.before_seconds_relative)
            if ret.assert_before_seconds is not None:
                ret.assert_before_seconds = min(ret.assert_before_seconds, s)
            else:
                ret.assert_before_seconds = s

    return ret


@dataclass
class SpendBundleAddInfo:
    cost: Optional[uint64]
    status: MempoolInclusionStatus
    removals: list[MempoolRemoveInfo]
    error: Optional[Err]


@dataclass
class NewPeakInfo:
    items: list[NewPeakItem]
    removals: list[MempoolRemoveInfo]


@dataclass
class NewPeakItem:
    transaction_id: bytes32
    spend_bundle: SpendBundle
    conds: SpendBundleConditions


# For block overhead cost calculation
QUOTE_BYTES = 2
QUOTE_EXECUTION_COST = 20


class MempoolManager:
    pool: Executor
    constants: ConsensusConstants
    seen_bundle_hashes: dict[bytes32, bytes32]
    get_coin_records: Callable[[Collection[bytes32]], Awaitable[list[CoinRecord]]]
    get_unspent_lineage_info_for_puzzle_hash: Callable[[bytes32], Awaitable[Optional[UnspentLineageInfo]]]
    nonzero_fee_minimum_fpc: int
    mempool_max_total_cost: int
    # a cache of MempoolItems that conflict with existing items in the pool
    _conflict_cache: ConflictTxCache
    # cache of MempoolItems with height conditions making them not valid yet
    _pending_cache: PendingTxCache
    seen_cache_size: int
    peak: Optional[BlockRecordProtocol]
    mempool: Mempool
    _worker_queue_size: int
    max_block_klvm_cost: uint64
    max_tx_klvm_cost: uint64

    def __init__(
        self,
        get_coin_records: Callable[[Collection[bytes32]], Awaitable[list[CoinRecord]]],
        get_unspent_lineage_info_for_puzzle_hash: Callable[[bytes32], Awaitable[Optional[UnspentLineageInfo]]],
        consensus_constants: ConsensusConstants,
        *,
        single_threaded: bool = False,
        max_tx_klvm_cost: Optional[uint64] = None,
    ):
        self.constants: ConsensusConstants = consensus_constants

        # Keep track of seen spend_bundles
        self.seen_bundle_hashes: dict[bytes32, bytes32] = {}

        self.get_coin_records = get_coin_records
        self.get_unspent_lineage_info_for_puzzle_hash = get_unspent_lineage_info_for_puzzle_hash

        # The fee per cost must be above this amount to consider the fee "nonzero", and thus able to kick out other
        # transactions. This prevents spam. This is equivalent to 0.055 XCK per block, or about 0.00005 XCK for two
        # spends.
        self.nonzero_fee_minimum_fpc = 5

        # We need to deduct the block overhead, which consists of the wrapping
        # quote opcode's bytes cost as well as its execution cost.
        BLOCK_OVERHEAD = QUOTE_BYTES * self.constants.COST_PER_BYTE + QUOTE_EXECUTION_COST

        self.max_block_klvm_cost = uint64(self.constants.MAX_BLOCK_COST_KLVM - BLOCK_OVERHEAD)
        self.max_tx_klvm_cost = (
            max_tx_klvm_cost if max_tx_klvm_cost is not None else uint64(self.constants.MAX_BLOCK_COST_KLVM // 2)
        )
        self.mempool_max_total_cost = int(self.constants.MAX_BLOCK_COST_KLVM * self.constants.MEMPOOL_BLOCK_BUFFER)

        # Transactions that were unable to enter mempool, used for retry. (they were invalid)
        self._conflict_cache = ConflictTxCache(self.constants.MAX_BLOCK_COST_KLVM * 1, 1000)
        self._pending_cache = PendingTxCache(self.constants.MAX_BLOCK_COST_KLVM * 1, 1000)
        self.seen_cache_size = 10000
        self._worker_queue_size = 0
        if single_threaded:
            self.pool = InlineExecutor()
        else:
            self.pool = ThreadPoolExecutor(max_workers=2)

        # The mempool will correspond to a certain peak
        self.peak: Optional[BlockRecordProtocol] = None
        self.fee_estimator: FeeEstimatorInterface = create_bitcoin_fee_estimator(self.max_block_klvm_cost)
        mempool_info = MempoolInfo(
            KLVMCost(uint64(self.mempool_max_total_cost)),
            FeeRate(uint64(self.nonzero_fee_minimum_fpc)),
            KLVMCost(uint64(self.max_block_klvm_cost)),
        )
        self.mempool: Mempool = Mempool(mempool_info, self.fee_estimator)

    def shut_down(self) -> None:
        self.pool.shutdown(wait=True)

    async def create_bundle_from_mempool(
        self,
        last_tb_header_hash: bytes32,
        get_unspent_lineage_info_for_puzzle_hash: Callable[[bytes32], Awaitable[Optional[UnspentLineageInfo]]],
        item_inclusion_filter: Optional[Callable[[bytes32], bool]] = None,
    ) -> Optional[tuple[SpendBundle, list[Coin]]]:
        """
        Returns aggregated spendbundle that can be used for creating new block,
        additions and removals in that spend_bundle
        """

        if self.peak is None or self.peak.header_hash != last_tb_header_hash:
            return None
        if item_inclusion_filter is None:

            def always(bundle_name: bytes32) -> bool:
                return True

            item_inclusion_filter = always
        return await self.mempool.create_bundle_from_mempool_items(
            item_inclusion_filter, get_unspent_lineage_info_for_puzzle_hash, self.constants, self.peak.height
        )

    def get_filter(self) -> bytes:
        all_transactions: set[bytes32] = set()
        byte_array_list = []
        for key in self.mempool.all_item_ids():
            if key not in all_transactions:
                all_transactions.add(key)
                byte_array_list.append(bytearray(key))

        tx_filter: PyBIP158 = PyBIP158(byte_array_list)
        return bytes(tx_filter.GetEncoded())

    def is_fee_enough(self, fees: uint64, cost: uint64) -> bool:
        """
        Determines whether any of the pools can accept a transaction with a given fees
        and cost.
        """
        if cost == 0:
            return False
        fees_per_cost = fees / cost
        if not self.mempool.at_full_capacity(cost):
            return True
        if fees_per_cost < self.nonzero_fee_minimum_fpc:
            return False
        min_fee_rate = self.mempool.get_min_fee_rate(cost)
        return min_fee_rate is not None and fees_per_cost > min_fee_rate

    def add_and_maybe_pop_seen(self, spend_name: bytes32) -> None:
        self.seen_bundle_hashes[spend_name] = spend_name
        while len(self.seen_bundle_hashes) > self.seen_cache_size:
            first_in = next(iter(self.seen_bundle_hashes.keys()))
            self.seen_bundle_hashes.pop(first_in)

    def seen(self, bundle_hash: bytes32) -> bool:
        """Return true if we saw this spendbundle recently"""
        return bundle_hash in self.seen_bundle_hashes

    def remove_seen(self, bundle_hash: bytes32) -> None:
        if bundle_hash in self.seen_bundle_hashes:
            self.seen_bundle_hashes.pop(bundle_hash)

    async def pre_validate_spendbundle(
        self, spend_bundle: SpendBundle, spend_bundle_id: Optional[bytes32] = None, bls_cache: Optional[BLSCache] = None
    ) -> SpendBundleConditions:
        """
        Errors are included within the cached_result.
        This runs in another process so we don't block the main thread
        """

        if spend_bundle.coin_spends == []:
            raise ValidationError(Err.INVALID_SPEND_BUNDLE, "Empty SpendBundle")

        assert self.peak is not None

        self._worker_queue_size += 1
        try:
            sbc, new_cache_entries, duration = await asyncio.get_running_loop().run_in_executor(
                self.pool,
                validate_klvm_and_signature,
                spend_bundle,
                self.max_tx_klvm_cost,
                self.constants,
                self.peak.height,
            )
        # validate_klvm_and_signature raises a TypeError with an error code
        except Exception as e:
            # Convert that to a ValidationError
            if len(e.args) > 0:
                error = Err(e.args[0])
                raise ValidationError(error)
            else:
                raise ValidationError(Err.UNKNOWN)  # pragma: no cover
        finally:
            self._worker_queue_size -= 1

        if bls_cache is not None:
            bls_cache.update(new_cache_entries)

        ret = NPCResult(None, sbc)

        if spend_bundle_id is None:
            spend_bundle_id = spend_bundle.name()

        log.log(
            logging.DEBUG if duration < 2 else logging.WARNING,
            f"pre_validate_spendbundle took {duration:0.4f} seconds "
            f"for {spend_bundle_id} (queue-size: {self._worker_queue_size})",
        )
        if ret.error is not None:
            raise ValidationError(Err(ret.error), "pre_validate_spendbundle failed")
        assert ret.conds is not None
        return ret.conds

    async def add_spend_bundle(
        self,
        new_spend: SpendBundle,
        conds: SpendBundleConditions,
        spend_name: bytes32,
        first_added_height: uint32,
        get_coin_records: Optional[Callable[[Collection[bytes32]], Awaitable[list[CoinRecord]]]] = None,
    ) -> SpendBundleAddInfo:
        """
        Validates and adds to mempool a new_spend with the given NPCResult, and spend_name, and the current mempool.
        The mempool should be locked during this call (blockchain lock). If there are mempool conflicts, the conflicting
        spends might be removed (if the new spend is a superset of the previous). Otherwise, the new spend might be
        added to the potential pool.

        Args:
            new_spend: spend bundle to validate and add
            conds: result of running the klvm transaction in a fake block
            spend_name: hash of the spend bundle data, passed in as an optimization

        Returns:
            Optional[uint64]: cost of the entire transaction, None iff status is FAILED
            MempoolInclusionStatus:  SUCCESS (should add to pool), FAILED (cannot add), and PENDING (can add later)
            list[MempoolRemoveInfo]: conflicting mempool items which were removed, if no Err
            Optional[Err]: Err is set iff status is FAILED
        """

        # Skip if already added
        existing_item = self.mempool.get_item_by_id(spend_name)
        if existing_item is not None:
            return SpendBundleAddInfo(existing_item.cost, MempoolInclusionStatus.SUCCESS, [], None)

        if get_coin_records is None:
            get_coin_records = self.get_coin_records
        err, item, remove_items = await self.validate_spend_bundle(
            new_spend,
            conds,
            spend_name,
            first_added_height,
            get_coin_records,
            self.get_unspent_lineage_info_for_puzzle_hash,
        )
        if err is None:
            # No error, immediately add to mempool, after removing conflicting TXs.
            assert item is not None
            conflict = self.mempool.remove_from_pool(remove_items, MempoolRemoveReason.CONFLICT)
            info = self.mempool.add_to_pool(item)
            if info.error is not None:
                return SpendBundleAddInfo(item.cost, MempoolInclusionStatus.FAILED, [], info.error)
            return SpendBundleAddInfo(item.cost, MempoolInclusionStatus.SUCCESS, [*info.removals, conflict], None)
        elif err is Err.MEMPOOL_CONFLICT and item is not None:
            # The transaction has a conflict with another item in the
            # mempool, put it aside and re-try it later
            self._conflict_cache.add(item)
            return SpendBundleAddInfo(item.cost, MempoolInclusionStatus.PENDING, [], err)
        elif item is not None:
            # The transasction has a height assertion and is not yet valid.
            # remember it to try it again later
            self._pending_cache.add(item)
            return SpendBundleAddInfo(item.cost, MempoolInclusionStatus.PENDING, [], err)
        else:
            # Cannot add to the mempool or pending pool.
            return SpendBundleAddInfo(None, MempoolInclusionStatus.FAILED, [], err)

    async def validate_spend_bundle(
        self,
        new_spend: SpendBundle,
        conds: SpendBundleConditions,
        spend_name: bytes32,
        first_added_height: uint32,
        get_coin_records: Callable[[Collection[bytes32]], Awaitable[list[CoinRecord]]],
        get_unspent_lineage_info_for_puzzle_hash: Callable[[bytes32], Awaitable[Optional[UnspentLineageInfo]]],
    ) -> tuple[Optional[Err], Optional[MempoolItem], list[bytes32]]:
        """
        Validates new_spend with the given NPCResult, and spend_name, and the current mempool. The mempool should
        be locked during this call (blockchain lock).

        Args:
            new_spend: spend bundle to validate
            conds: result of running the klvm transaction in a fake block
            spend_name: hash of the spend bundle data, passed in as an optimization
            first_added_height: The block height that `new_spend`  first entered this node's mempool.
                Used to estimate how long a spend has taken to be included on the chain.
                This value could differ node to node. Not preserved across full_node restarts.

        Returns:
            Optional[Err]: Err is set if we cannot add to the mempool, None if we will immediately add to mempool
            Optional[MempoolItem]: the item to add (to mempool or pending pool)
            list[bytes32]: conflicting mempool items to remove, if no Err
        """
        start_time = time.time()
        if self.peak is None:
            return Err.MEMPOOL_NOT_INITIALIZED, None, []

        cost = conds.cost

        removal_names: set[bytes32] = set()
        additions_dict: dict[bytes32, Coin] = {}
        addition_amount: int = 0
        # Map of coin ID to eligibility information
        eligibility_and_additions: dict[bytes32, EligibilityAndAdditions] = {}
        non_eligible_coin_ids: list[bytes32] = []
        for spend in conds.spends:
            coin_id = bytes32(spend.coin_id)
            removal_names.add(coin_id)
            spend_additions = []
            for puzzle_hash, amount, _ in spend.create_coin:
                child_coin = Coin(coin_id, puzzle_hash, uint64(amount))
                spend_additions.append(child_coin)
                additions_dict[child_coin.name()] = child_coin
                addition_amount += child_coin.amount
            is_eligible_for_dedup = bool(spend.flags & ELIGIBLE_FOR_DEDUP)
            is_eligible_for_ff = bool(spend.flags & ELIGIBLE_FOR_FF)
            eligibility_and_additions[coin_id] = EligibilityAndAdditions(
                is_eligible_for_dedup=is_eligible_for_dedup,
                spend_additions=spend_additions,
                ff_puzzle_hash=bytes32(spend.puzzle_hash) if is_eligible_for_ff else None,
            )
        removal_names_from_coin_spends: set[bytes32] = set()
        fast_forward_coin_ids: set[bytes32] = set()
        bundle_coin_spends: dict[bytes32, BundleCoinSpend] = {}
        for coin_spend in new_spend.coin_spends:
            coin_id = coin_spend.coin.name()
            removal_names_from_coin_spends.add(coin_id)
            eligibility_info = eligibility_and_additions.get(
                coin_id,
                EligibilityAndAdditions(is_eligible_for_dedup=False, spend_additions=[], ff_puzzle_hash=None),
            )
            mark_as_fast_forward = eligibility_info.ff_puzzle_hash is not None and supports_fast_forward(coin_spend)
            if mark_as_fast_forward:
                # Make sure the fast forward spend still has a version that is
                # still unspent, because if the singleton has been melted, the
                # fast forward spend will never become valid.
                assert eligibility_info.ff_puzzle_hash is not None
                if await get_unspent_lineage_info_for_puzzle_hash(eligibility_info.ff_puzzle_hash) is None:
                    return Err.DOUBLE_SPEND, None, []
                fast_forward_coin_ids.add(coin_id)
            # We are now able to check eligibility of both dedup and fast forward
            if not (eligibility_info.is_eligible_for_dedup or mark_as_fast_forward):
                non_eligible_coin_ids.append(coin_id)
            bundle_coin_spends[coin_id] = BundleCoinSpend(
                coin_spend=coin_spend,
                eligible_for_dedup=eligibility_info.is_eligible_for_dedup,
                eligible_for_fast_forward=mark_as_fast_forward,
                additions=eligibility_info.spend_additions,
            )

        if removal_names != removal_names_from_coin_spends:
            # If you reach here it's probably because your program reveal doesn't match the coin's puzzle hash
            return Err.INVALID_SPEND_BUNDLE, None, []

        # fast forward spends are only allowed when bundled with other, non-FF, spends
        # in order to evict an FF spend, it must be associated with a normal
        # spend that can be included in a block or invalidated some other way
        if all([s.eligible_for_fast_forward for s in bundle_coin_spends.values()]):
            return Err.INVALID_SPEND_BUNDLE, None, []

        removal_record_dict: dict[bytes32, CoinRecord] = {}
        removal_amount: int = 0
        removal_records = await get_coin_records(removal_names)
        for record in removal_records:
            removal_record_dict[record.coin.name()] = record

        for name in removal_names:
            if name not in removal_record_dict and name not in additions_dict:
                return Err.UNKNOWN_UNSPENT, None, []
            if name in additions_dict:
                removal_coin = additions_dict[name]
                # The timestamp and block-height of this coin being spent needs
                # to be consistent with what we use to check time-lock
                # conditions (below). All spends (including ephemeral coins) are
                # spent simultaneously. Ephemeral coins with an
                # ASSERT_SECONDS_RELATIVE 0 condition are still OK to spend in
                # the same block.
                assert self.peak.timestamp is not None
                removal_record = CoinRecord(
                    removal_coin,
                    uint32(self.peak.height + 1),
                    uint32(0),
                    False,
                    self.peak.timestamp,
                )
                removal_record_dict[name] = removal_record
            else:
                removal_record = removal_record_dict[name]
            removal_amount += removal_record.coin.amount

        fees = uint64(removal_amount - addition_amount)

        if cost == 0:
            return Err.UNKNOWN, None, []

        if cost > self.max_tx_klvm_cost:
            return Err.BLOCK_COST_EXCEEDS_MAX, None, []

        # this is not very likely to happen, but it's here to ensure SQLite
        # never runs out of precision in its computation of fees.
        # sqlite's integers are signed int64, so the max value they can
        # represent is 2^63-1
        if fees > MEMPOOL_ITEM_FEE_LIMIT or SQLITE_INT_MAX - self.mempool.total_mempool_fees() <= fees:
            return Err.INVALID_BLOCK_FEE_AMOUNT, None, []

        fees_per_cost: float = fees / cost
        # If pool is at capacity check the fee, if not then accept even without the fee
        if self.mempool.at_full_capacity(cost):
            if fees_per_cost < self.nonzero_fee_minimum_fpc:
                return Err.INVALID_FEE_TOO_CLOSE_TO_ZERO, None, []
            min_fee_rate = self.mempool.get_min_fee_rate(cost)
            if min_fee_rate is None:
                return Err.INVALID_COST_RESULT, None, []
            if fees_per_cost <= min_fee_rate:
                return Err.INVALID_FEE_LOW_FEE, None, []

        # Check removals against UnspentDB + DiffStore + Mempool + SpendBundle
        # Use this information later when constructing a block
        fail_reason, conflicts = self.check_removals(non_eligible_coin_ids, removal_record_dict, fast_forward_coin_ids)

        # If we have a mempool conflict, continue, since we still want to keep around the TX in the pending pool.
        if fail_reason is not None and fail_reason is not Err.MEMPOOL_CONFLICT:
            return fail_reason, None, []

        # Verify conditions, create hash_key list for aggsig check
        for spend in conds.spends:
            coin_record: CoinRecord = removal_record_dict[bytes32(spend.coin_id)]
            # Check that the revealed removal puzzles actually match the puzzle hash
            if spend.puzzle_hash != coin_record.coin.puzzle_hash:
                log.warning("Mempool rejecting transaction because of wrong puzzle_hash")
                log.warning(f"{spend.puzzle_hash.hex()} != {coin_record.coin.puzzle_hash.hex()}")
                return Err.WRONG_PUZZLE_HASH, None, []

        # the height and time we pass in here represent the previous transaction
        # block's height and timestamp. In the mempool, the most recent peak
        # block we've received will be the previous transaction block, from the
        # point-of-view of the next block to be farmed. Therefore we pass in the
        # current peak's height and timestamp
        assert self.peak.timestamp is not None
        tl_error: Optional[Err] = mempool_check_time_locks(
            removal_record_dict,
            conds,
            self.peak.height,
            self.peak.timestamp,
        )

        timelocks: TimelockConditions = compute_assert_height(removal_record_dict, conds)

        if timelocks.assert_before_height is not None and timelocks.assert_before_height <= timelocks.assert_height:
            # returning None as the "potential" means it failed. We won't store it
            # in the pending cache
            return Err.IMPOSSIBLE_HEIGHT_ABSOLUTE_CONSTRAINTS, None, []  # MempoolInclusionStatus.FAILED
        if timelocks.assert_before_seconds is not None and timelocks.assert_before_seconds <= timelocks.assert_seconds:
            return Err.IMPOSSIBLE_SECONDS_ABSOLUTE_CONSTRAINTS, None, []  # MempoolInclusionStatus.FAILED

        potential = MempoolItem(
            new_spend,
            uint64(fees),
            conds,
            spend_name,
            first_added_height,
            timelocks.assert_height,
            timelocks.assert_before_height,
            timelocks.assert_before_seconds,
            bundle_coin_spends,
        )

        if tl_error:
            if tl_error is Err.ASSERT_HEIGHT_ABSOLUTE_FAILED or tl_error is Err.ASSERT_HEIGHT_RELATIVE_FAILED:
                return tl_error, potential, []  # MempoolInclusionStatus.PENDING
            else:
                return tl_error, None, []  # MempoolInclusionStatus.FAILED

        if fail_reason is Err.MEMPOOL_CONFLICT:
            log.debug(f"Replace attempted. number of MempoolItems: {len(conflicts)}")
            if not can_replace(conflicts, removal_names, potential):
                return Err.MEMPOOL_CONFLICT, potential, []

        duration = time.time() - start_time

        log.log(
            logging.DEBUG if duration < 2 else logging.WARNING,
            f"add_spendbundle {spend_name} took {duration:0.2f} seconds. "
            f"Cost: {cost} ({round(100.0 * cost / self.constants.MAX_BLOCK_COST_KLVM, 3)}% of max block cost)",
        )

        return None, potential, [item.name for item in conflicts]

    def check_removals(
        self,
        non_eligible_coin_ids: list[bytes32],
        removals: dict[bytes32, CoinRecord],
        fast_forward_coin_ids: set[bytes32],
    ) -> tuple[Optional[Err], list[MempoolItem]]:
        """
        This function checks for double spends, unknown spends and conflicting transactions in mempool.
        Returns Error (if any), the set of existing MempoolItems with conflicting spends (if any).
        Note that additions are not checked for duplicates, because having duplicate additions requires also
        having duplicate removals.
        """
        assert self.peak is not None
        # 1. Checks if it's been spent already
        for record in removals.values():
            if record.spent:
                # Only consider it a double spend if this is not a fast forward
                if record.name not in fast_forward_coin_ids:
                    return Err.DOUBLE_SPEND, []
        # 2. Checks if there's a mempool conflict
        # Only consider conflicts if the coin is not eligible for deduplication
        conflicts = self.mempool.get_items_by_coin_ids(non_eligible_coin_ids)
        if len(conflicts) > 0:
            return Err.MEMPOOL_CONFLICT, conflicts
        # 5. If coins can be spent return list of unspents as we see them in local storage
        return None, []

    def get_spendbundle(self, bundle_hash: bytes32) -> Optional[SpendBundle]:
        """Returns a full SpendBundle if it's inside one the mempools"""
        item: Optional[MempoolItem] = self.mempool.get_item_by_id(bundle_hash)
        if item is not None:
            return item.spend_bundle
        return None

    def get_mempool_item(self, bundle_hash: bytes32, include_pending: bool = False) -> Optional[MempoolItem]:
        """
        Returns the MempoolItem in the mempool that matches the provided spend bundle hash (id)
        or None if not found.

        If include_pending is specified, also check the PENDING cache.
        """
        item = self.mempool.get_item_by_id(bundle_hash)
        if not item and include_pending:
            # no async lock needed since we're not mutating the pending_cache
            item = self._pending_cache.get(bundle_hash)
        if not item and include_pending:
            item = self._conflict_cache.get(bundle_hash)

        return item

    async def new_peak(
        self, new_peak: Optional[BlockRecordProtocol], spent_coins: Optional[list[bytes32]]
    ) -> NewPeakInfo:
        """
        Called when a new peak is available, we try to recreate a mempool for the new tip.
        new_peak should always be the most recent *transaction* block of the chain. Since
        the mempool cannot traverse the chain to find the most recent transaction block,
        we wouldn't be able to detect, and correctly update the mempool, if we saw a
        non-transaction block on a fork. self.peak must always be set to a transaction
        block.
        """
        if new_peak is None:
            return NewPeakInfo([], [])
        # we're only interested in transaction blocks
        if new_peak.is_transaction_block is False:
            return NewPeakInfo([], [])
        if self.peak == new_peak:
            return NewPeakInfo([], [])
        assert new_peak.timestamp is not None
        self.fee_estimator.new_block_height(new_peak.height)
        included_items: list[MempoolItemInfo] = []

        expired = self.mempool.new_tx_block(new_peak.height, new_peak.timestamp)
        mempool_item_removals: list[MempoolRemoveInfo] = [expired]

        use_optimization: bool = self.peak is not None and new_peak.prev_transaction_block_hash == self.peak.header_hash
        self.peak = new_peak

        if use_optimization and spent_coins is not None:
            # We don't reinitialize a mempool, just kick removed items
            # transactions in the mempool may be spending multiple coins,
            # when looking up transactions by all coin IDs, we're likely to
            # find the same transaction multiple times. We put them in a set
            # to deduplicate
            spendbundle_ids_to_remove: set[bytes32] = set()
            for spend in spent_coins:
                items = self.mempool.get_items_by_coin_id(spend)
                for item in items:
                    included_items.append(MempoolItemInfo(item.cost, item.fee, item.height_added_to_mempool))
                    self.remove_seen(item.name)
                    spendbundle_ids_to_remove.add(item.name)
            mempool_item_removals.append(
                self.mempool.remove_from_pool(list(spendbundle_ids_to_remove), MempoolRemoveReason.BLOCK_INCLUSION)
            )
        else:
            log.warning(
                "updating the mempool using the slow-path. "
                f"peak: {self.peak.header_hash.hex()} "
                f"new-peak-prev: {new_peak.prev_transaction_block_hash} "
                f"coins: {'not set' if spent_coins is None else 'set'}"
            )
            old_pool = self.mempool
            self.mempool = Mempool(old_pool.mempool_info, old_pool.fee_estimator)
            self.seen_bundle_hashes = {}

            # in order to make this a bit quicker, we look-up all the spends in
            # a single query, rather than one at a time.
            coin_records: dict[bytes32, CoinRecord] = {}

            removals: set[bytes32] = set()
            for item in old_pool.all_items():
                for s in item.spend_bundle.coin_spends:
                    removals.add(s.coin.name())

            for record in await self.get_coin_records(removals):
                name = record.coin.name()
                coin_records[name] = record

            async def local_get_coin_records(names: Collection[bytes32]) -> list[CoinRecord]:
                ret: list[CoinRecord] = []
                for name in names:
                    r = coin_records.get(name)
                    if r is not None:
                        ret.append(r)
                return ret

            for item in old_pool.all_items():
                info = await self.add_spend_bundle(
                    item.spend_bundle,
                    item.conds,
                    item.spend_bundle_name,
                    item.height_added_to_mempool,
                    local_get_coin_records,
                )
                # Only add to `seen` if inclusion worked, so it can be resubmitted in case of a reorg
                if info.status == MempoolInclusionStatus.SUCCESS:
                    self.add_and_maybe_pop_seen(item.spend_bundle_name)
                # If the spend bundle was confirmed or conflicting (can no longer be in mempool), it won't be
                # successfully added to the new mempool.
                if info.status == MempoolInclusionStatus.FAILED and info.error == Err.DOUBLE_SPEND:
                    # Item was in mempool, but after the new block it's a double spend.
                    # Item is most likely included in the block.
                    included_items.append(MempoolItemInfo(item.cost, item.fee, item.height_added_to_mempool))

        potential_txs = self._pending_cache.drain(new_peak.height)
        potential_txs.update(self._conflict_cache.drain())
        txs_added = []
        for item in potential_txs.values():
            info = await self.add_spend_bundle(
                item.spend_bundle,
                item.conds,
                item.spend_bundle_name,
                item.height_added_to_mempool,
                self.get_coin_records,
            )
            if info.status == MempoolInclusionStatus.SUCCESS:
                txs_added.append(NewPeakItem(item.spend_bundle_name, item.spend_bundle, item.conds))
            mempool_item_removals.extend(info.removals)
        log.info(
            f"Size of mempool: {self.mempool.size()} spends, "
            f"cost: {self.mempool.total_mempool_cost()} "
            f"minimum fee rate (in FPC) to get in for 5M cost tx: {self.mempool.get_min_fee_rate(5000000)}"
        )
        self.mempool.fee_estimator.new_block(FeeBlockInfo(new_peak.height, included_items))
        return NewPeakInfo(txs_added, mempool_item_removals)

    def get_items_not_in_filter(self, mempool_filter: PyBIP158, limit: int = 100) -> list[SpendBundle]:
        items: list[SpendBundle] = []

        assert limit > 0

        # Send 100 with the highest fee per cost
        for item in self.mempool.items_by_feerate():
            if len(items) >= limit:
                return items
            if mempool_filter.Match(bytearray(item.spend_bundle_name)):
                continue
            items.append(item.spend_bundle)

        return items


T = TypeVar("T", uint32, uint64)


def optional_min(a: Optional[T], b: Optional[T]) -> Optional[T]:
    return min((v for v in [a, b] if v is not None), default=None)


def optional_max(a: Optional[T], b: Optional[T]) -> Optional[T]:
    return max((v for v in [a, b] if v is not None), default=None)


def can_replace(
    conflicting_items: list[MempoolItem],
    removal_names: set[bytes32],
    new_item: MempoolItem,
) -> bool:
    """
    This function implements the mempool replacement rules. Given a Mempool item
    we're attempting to insert into the mempool (new_item) and the set of existing
    mempool items that conflict with it, this function answers the question whether
    the existing items can be replaced by the new one. The removals parameter are
    the coin IDs the new mempool item is spending.
    """

    conflicting_fees = 0
    conflicting_cost = 0
    assert_height: Optional[uint32] = None
    assert_before_height: Optional[uint32] = None
    assert_before_seconds: Optional[uint64] = None
    for item in conflicting_items:
        conflicting_fees += item.fee
        conflicting_cost += item.cost

        # All coins spent in all conflicting items must also be spent in the new item. (superset rule). This is
        # important because otherwise there exists an attack. A user spends coin A. An attacker replaces the
        # bundle with AB with a higher fee. An attacker then replaces the bundle with just B with a higher
        # fee than AB therefore kicking out A altogether. The better way to solve this would be to keep a cache
        # of booted transactions like A, and retry them after they get removed from mempool due to a conflict.
        for coin in item.removals:
            if coin.name() not in removal_names:
                log.debug(f"Rejecting conflicting tx as it does not spend conflicting coin {coin.name()}")
                return False

        assert_height = optional_max(assert_height, item.assert_height)
        assert_before_height = optional_min(assert_before_height, item.assert_before_height)
        assert_before_seconds = optional_min(assert_before_seconds, item.assert_before_seconds)

    # New item must have higher fee per cost
    conflicting_fees_per_cost = conflicting_fees / conflicting_cost
    if new_item.fee_per_cost <= conflicting_fees_per_cost:
        log.debug(
            f"Rejecting conflicting tx due to not increasing fees per cost "
            f"({new_item.fee_per_cost} <= {conflicting_fees_per_cost})"
        )
        return False

    # New item must increase the total fee at least by a certain amount
    fee_increase = new_item.fee - conflicting_fees
    if fee_increase < MEMPOOL_MIN_FEE_INCREASE:
        log.debug(f"Rejecting conflicting tx due to low fee increase ({fee_increase})")
        return False

    # New item may not have a different effective height/time lock (time-lock rule)
    if new_item.assert_height != assert_height:
        log.debug(
            "Rejecting conflicting tx due to changing ASSERT_HEIGHT constraints %s -> %s",
            assert_height,
            new_item.assert_height,
        )
        return False

    if new_item.assert_before_height != assert_before_height:
        log.debug(
            "Rejecting conflicting tx due to changing ASSERT_BEFORE_HEIGHT constraints %s -> %s",
            assert_before_height,
            new_item.assert_before_height,
        )
        return False

    if new_item.assert_before_seconds != assert_before_seconds:
        log.debug(
            "Rejecting conflicting tx due to changing ASSERT_BEFORE_SECONDS constraints %s -> %s",
            assert_before_seconds,
            new_item.assert_before_seconds,
        )
        return False

    log.info(f"Replacing conflicting tx in mempool. New tx fee: {new_item.fee}, old tx fees: {conflicting_fees}")
    return True



from __future__ import annotations

import dataclasses
from collections.abc import Iterator
from typing import Optional, Union

from chik_rs import G2Element

from chik.types.blockchain_format.coin import Coin, coin_as_list
from chik.types.blockchain_format.program import INFINITE_COST, Program
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.coin_spend import make_spend
from chik.types.condition_opcodes import ConditionOpcode
from chik.util.condition_tools import conditions_dict_for_solution
from chik.wallet.lineage_proof import LineageProof
from chik.wallet.puzzles.load_klvm import load_klvm_maybe_recompile
from chik.wallet.uncurried_puzzle import UncurriedPuzzle
from chik.wallet.wallet_spend_bundle import WalletSpendBundle

NULL_SIGNATURE = G2Element()

ANYONE_CAN_SPEND_PUZZLE = Program.to(1)  # simply return the conditions
CAT_MOD = load_klvm_maybe_recompile("cat_v2.clsp", package_or_requirement="chik.wallet.cat_wallet.puzzles")
CAT_MOD_HASH = CAT_MOD.get_tree_hash()
CAT_MOD_HASH_HASH: bytes32 = Program.to(CAT_MOD_HASH).get_tree_hash()


def empty_program() -> Program:
    return Program.to([])


# information needed to spend a cc
@dataclasses.dataclass
class SpendableCAT:
    coin: Coin
    limitations_program_hash: bytes32
    inner_puzzle: Program
    inner_solution: Program
    limitations_solution: Program = dataclasses.field(default_factory=empty_program)
    lineage_proof: LineageProof = LineageProof()
    extra_delta: int = 0
    limitations_program_reveal: Program = dataclasses.field(default_factory=empty_program)


def match_cat_puzzle(puzzle: UncurriedPuzzle) -> Optional[Iterator[Program]]:
    """
    Given the curried puzzle and args, test if it's a CAT and,
    if it is, return the curried arguments
    """
    if puzzle.mod == CAT_MOD:
        ret: Iterator[Program] = puzzle.args.as_iter()
        return ret
    else:
        return None


def get_innerpuzzle_from_puzzle(puzzle: Program) -> Program:
    mod, curried_args = puzzle.uncurry()
    if mod == CAT_MOD:
        return curried_args.at("rrf")
    else:
        raise ValueError("Not a CAT puzzle")


def construct_cat_puzzle(
    mod_code: Program,
    limitations_program_hash: bytes32,
    inner_puzzle_or_hash: Union[Program, bytes32],
    mod_code_hash: Optional[bytes32] = None,
) -> Program:
    """
    Given an inner puzzle and a tail hash, calculate a puzzle program for a specific cc.
    We can also receive an inner puzzle hash instead, which wouldn't calculate a valid
    puzzle, but that can be useful if calling `.get_tree_hash_precalc()` on it."
    """
    if mod_code_hash is None:
        mod_code_hash = mod_code.get_tree_hash()
    return mod_code.curry(mod_code_hash, limitations_program_hash, inner_puzzle_or_hash)


def subtotals_for_deltas(deltas: list[int]) -> list[int]:
    """
    Given a list of deltas corresponding to input coins, create the "subtotals" list
    needed in solutions spending those coins.
    """

    subtotals = []
    subtotal = 0

    for delta in deltas:
        subtotals.append(subtotal)
        subtotal += delta

    # tweak the subtotals so the smallest value is 0
    subtotal_offset = min(subtotals)
    subtotals = [_ - subtotal_offset for _ in subtotals]
    return subtotals


def next_info_for_spendable_cat(spendable_cat: SpendableCAT) -> Program:
    c = spendable_cat.coin
    list = [c.parent_coin_info, spendable_cat.inner_puzzle.get_tree_hash(), c.amount]
    return Program.to(list)


# This should probably return UnsignedSpendBundle if that type ever exists
def unsigned_spend_bundle_for_spendable_cats(
    mod_code: Program, spendable_cat_list: list[SpendableCAT]
) -> WalletSpendBundle:
    """
    Given a list of `SpendableCAT` objects, create a `WalletSpendBundle` that spends all those coins.
    Note that no signing is done here, so it falls on the caller to sign the resultant bundle.
    """

    N = len(spendable_cat_list)

    # figure out what the deltas are by running the inner puzzles & solutions
    deltas: list[int] = []
    for spend_info in spendable_cat_list:
        conditions = conditions_dict_for_solution(spend_info.inner_puzzle, spend_info.inner_solution, INFINITE_COST)
        total = spend_info.extra_delta * -1
        for _ in conditions.get(ConditionOpcode.CREATE_COIN, []):
            if _.vars[1] != b"\x8f":  # -113 in bytes
                total += Program.to(_.vars[1]).as_int()
        deltas.append(spend_info.coin.amount - total)

    if sum(deltas) != 0:
        raise ValueError("input and output amounts don't match")

    subtotals = subtotals_for_deltas(deltas)

    infos_for_next = []
    infos_for_me = []
    ids = []
    for _ in spendable_cat_list:
        infos_for_next.append(next_info_for_spendable_cat(_))
        infos_for_me.append(Program.to(coin_as_list(_.coin)))
        ids.append(_.coin.name())

    coin_spends = []
    for index in range(N):
        spend_info = spendable_cat_list[index]

        puzzle_reveal = construct_cat_puzzle(mod_code, spend_info.limitations_program_hash, spend_info.inner_puzzle)

        prev_index = (index - 1) % N
        next_index = (index + 1) % N
        prev_id = ids[prev_index]
        my_info = infos_for_me[index]
        next_info = infos_for_next[next_index]

        solution = [
            spend_info.inner_solution,
            spend_info.lineage_proof.to_program(),
            prev_id,
            my_info,
            next_info,
            subtotals[index],
            spend_info.extra_delta,
        ]
        coin_spend = make_spend(spend_info.coin, puzzle_reveal, Program.to(solution))
        coin_spends.append(coin_spend)

    return WalletSpendBundle(coin_spends, NULL_SIGNATURE)



from __future__ import annotations

import pytest
from chik_rs import AugSchemeMPL
from klvm.casts import int_to_bytes

from chik import __version__
from chik._tests.blockchain.blockchain_test_utils import _validate_and_add_block
from chik._tests.connection_utils import connect_and_get_peer
from chik._tests.util.rpc import validate_get_routes
from chik._tests.util.time_out_assert import time_out_assert
from chik.consensus.block_record import BlockRecord
from chik.consensus.pot_iterations import is_overflow_block
from chik.full_node.signage_point import SignagePoint
from chik.protocols import full_node_protocol
from chik.rpc.full_node_rpc_api import get_average_block_time, get_nearest_transaction_block
from chik.rpc.full_node_rpc_client import FullNodeRpcClient
from chik.server.outbound_message import NodeType
from chik.simulator.add_blocks_in_batches import add_blocks_in_batches
from chik.simulator.block_tools import get_signage_point
from chik.simulator.simulator_protocol import FarmNewBlockProtocol, ReorgProtocol
from chik.simulator.wallet_tools import WalletTool
from chik.types.blockchain_format.coin import Coin
from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.coin_spend import compute_additions
from chik.types.condition_opcodes import ConditionOpcode
from chik.types.condition_with_args import ConditionWithArgs
from chik.types.full_block import FullBlock
from chik.types.unfinished_block import UnfinishedBlock
from chik.util.hash import std_hash
from chik.util.ints import uint8
from chik.wallet.wallet_spend_bundle import WalletSpendBundle


@pytest.mark.anyio
async def test1(two_nodes_sim_and_wallets_services, self_hostname, consensus_mode):
    num_blocks = 5
    nodes, _, bt = two_nodes_sim_and_wallets_services
    full_node_service_1, full_node_service_2 = nodes
    full_node_api_1 = full_node_service_1._api
    full_node_api_2 = full_node_service_2._api
    server_2 = full_node_api_2.full_node.server

    async with FullNodeRpcClient.create_as_context(
        self_hostname,
        full_node_service_1.rpc_server.listen_port,
        full_node_service_1.root_path,
        full_node_service_1.config,
    ) as client:
        await validate_get_routes(client, full_node_service_1.rpc_server.rpc_api)
        state = await client.get_blockchain_state()
        assert state["peak"] is None
        assert not state["sync"]["sync_mode"]
        assert state["difficulty"] > 0
        assert state["sub_slot_iters"] > 0

        blocks = bt.get_consecutive_blocks(num_blocks)
        blocks = bt.get_consecutive_blocks(num_blocks, block_list_input=blocks, guarantee_transaction_block=True)

        assert len(await client.get_unfinished_block_headers()) == 0
        assert len(await client.get_block_records(0, 100)) == 0
        for block in blocks:
            if is_overflow_block(bt.constants, block.reward_chain_block.signage_point_index):
                finished_ss = block.finished_sub_slots[:-1]
            else:
                finished_ss = block.finished_sub_slots

            unf = UnfinishedBlock(
                finished_ss,
                block.reward_chain_block.get_unfinished(),
                block.challenge_chain_sp_proof,
                block.reward_chain_sp_proof,
                block.foliage,
                block.foliage_transaction_block,
                block.transactions_info,
                block.transactions_generator,
                [],
            )
            await full_node_api_1.full_node.add_unfinished_block(unf, None)
            await full_node_api_1.full_node.add_block(block, None)

        assert len(await client.get_unfinished_block_headers()) > 0
        assert len(await client.get_all_block(0, 2)) == 2
        state = await client.get_blockchain_state()

        block = await client.get_block(state["peak"].header_hash)
        assert block == blocks[-1]
        assert (await client.get_block(bytes([1] * 32))) is None

        assert (await client.get_block_record_by_height(2)).header_hash == blocks[2].header_hash

        assert len(await client.get_block_records(0, 100)) == num_blocks * 2

        assert (await client.get_block_record_by_height(100)) is None

        # NOTE: indexing and hard coded values below depend on the ordering
        included_reward_coins = sorted(blocks[-1].get_included_reward_coins(), key=lambda c: c.amount)

        ph = included_reward_coins[0].puzzle_hash
        coins = await client.get_coin_records_by_puzzle_hash(ph)
        print(coins)
        assert len(coins) >= 1

        pid = included_reward_coins[0].parent_coin_info
        pid_2 = included_reward_coins[1].parent_coin_info
        coins = await client.get_coin_records_by_parent_ids([pid, pid_2])
        print(coins)
        assert len(coins) == 2

        name = included_reward_coins[0].name()
        name_2 = included_reward_coins[1].name()
        coins = await client.get_coin_records_by_names([name, name_2])
        print(coins)
        assert len(coins) == 2

        additions, removals = await client.get_additions_and_removals(blocks[-1].header_hash)
        assert len(additions) >= 2 and len(removals) == 0

        wallet = WalletTool(full_node_api_1.full_node.constants)
        wallet_receiver = WalletTool(full_node_api_1.full_node.constants, AugSchemeMPL.key_gen(std_hash(b"123123")))
        ph = wallet.get_new_puzzlehash()
        ph_2 = wallet.get_new_puzzlehash()
        ph_receiver = wallet_receiver.get_new_puzzlehash()

        assert len(await client.get_coin_records_by_puzzle_hash(ph)) == 0
        assert len(await client.get_coin_records_by_puzzle_hash(ph_receiver)) == 0
        blocks = bt.get_consecutive_blocks(
            2,
            block_list_input=blocks,
            guarantee_transaction_block=True,
            farmer_reward_puzzle_hash=ph,
            pool_reward_puzzle_hash=ph,
        )
        for block in blocks[-2:]:
            await full_node_api_1.full_node.add_block(block)
        assert len(await client.get_coin_records_by_puzzle_hash(ph)) == 2
        assert len(await client.get_coin_records_by_puzzle_hash(ph_receiver)) == 0

        # NOTE: indexing and hard coded values below depend on the ordering
        included_reward_coins = sorted(blocks[-1].get_included_reward_coins(), key=lambda c: c.amount)

        coin_to_spend = included_reward_coins[0]

        spend_bundle = wallet.generate_signed_transaction(coin_to_spend.amount, ph_receiver, coin_to_spend)

        assert len(await client.get_all_mempool_items()) == 0
        assert len(await client.get_all_mempool_tx_ids()) == 0
        assert (await client.get_mempool_item_by_tx_id(spend_bundle.name())) is None
        assert (await client.get_mempool_item_by_tx_id(spend_bundle.name(), False)) is None

        await client.push_tx(spend_bundle)
        coin = spend_bundle.additions()[0]

        assert len(await client.get_all_mempool_items()) == 1
        assert len(await client.get_all_mempool_tx_ids()) == 1
        assert (
            WalletSpendBundle.from_json_dict(
                next(iter((await client.get_all_mempool_items()).values()))["spend_bundle"]
            )
            == spend_bundle
        )
        assert (await client.get_all_mempool_tx_ids())[0] == spend_bundle.name()
        assert (
            WalletSpendBundle.from_json_dict(
                (await client.get_mempool_item_by_tx_id(spend_bundle.name()))["spend_bundle"]
            )
            == spend_bundle
        )
        assert (await client.get_coin_record_by_name(coin.name())) is None

        # Verify that the include_pending arg to get_mempool_item_by_tx_id works
        coin_to_spend_pending = included_reward_coins[1]
        ahr = ConditionOpcode.ASSERT_HEIGHT_RELATIVE  # to force pending/potential
        condition_dic = {ahr: [ConditionWithArgs(ahr, [int_to_bytes(100)])]}
        spend_bundle_pending = wallet.generate_signed_transaction(
            coin_to_spend_pending.amount,
            ph_receiver,
            coin_to_spend_pending,
            condition_dic=condition_dic,
        )
        await client.push_tx(spend_bundle_pending)
        assert (
            await client.get_mempool_item_by_tx_id(spend_bundle_pending.name(), False)
        ) is None  # not strictly in the mempool
        assert (
            WalletSpendBundle.from_json_dict(
                (await client.get_mempool_item_by_tx_id(spend_bundle_pending.name(), True))["spend_bundle"]
            )
            == spend_bundle_pending  # pending entry into mempool, so include_pending fetches
        )

        await full_node_api_1.farm_new_transaction_block(FarmNewBlockProtocol(ph_2))

        coin_record = await client.get_coin_record_by_name(coin.name())
        assert coin_record.coin == coin
        assert coin in compute_additions(
            await client.get_puzzle_and_solution(coin.parent_coin_info, coin_record.confirmed_block_index)
        )

        assert len(await client.get_coin_records_by_puzzle_hash(ph_receiver)) == 1
        assert len(list(filter(lambda cr: not cr.spent, (await client.get_coin_records_by_puzzle_hash(ph))))) == 3
        assert len(await client.get_coin_records_by_puzzle_hashes([ph_receiver, ph])) == 5
        assert len(await client.get_coin_records_by_puzzle_hash(ph, False)) == 3
        assert len(await client.get_coin_records_by_puzzle_hash(ph, True)) == 4

        assert len(await client.get_coin_records_by_puzzle_hash(ph, True, 0, 100)) == 4
        assert len(await client.get_coin_records_by_puzzle_hash(ph, True, 50, 100)) == 0
        assert len(await client.get_coin_records_by_puzzle_hash(ph, True, 0, blocks[-1].height + 1)) == 2
        assert len(await client.get_coin_records_by_puzzle_hash(ph, True, 0, 1)) == 0

        coin_records = await client.get_coin_records_by_puzzle_hash(ph, False)

        coin_spends = []

        # Spend 3 coins using standard transaction
        for i in range(3):
            spend_bundle = wallet.generate_signed_transaction(
                coin_records[i].coin.amount, ph_receiver, coin_records[i].coin
            )
            await client.push_tx(spend_bundle)
            coin_spends += spend_bundle.coin_spends
            await time_out_assert(
                5, full_node_api_1.full_node.mempool_manager.get_spendbundle, spend_bundle, spend_bundle.name()
            )

        await full_node_api_1.farm_new_transaction_block(FarmNewBlockProtocol(ph_2))
        block: FullBlock = (await full_node_api_1.get_all_full_blocks())[-1]

        # since the hard fork, we no longer compress blocks using
        # block references anymore
        assert block.transactions_generator_ref_list == []

        block_spends = await client.get_block_spends(block.header_hash)

        assert len(block_spends) == 3
        assert sorted(block_spends, key=str) == sorted(coin_spends, key=str)

        block_spends_with_conditions = await client.get_block_spends_with_conditions(block.header_hash)

        assert len(block_spends_with_conditions) == 3

        block_spends_with_conditions = sorted(block_spends_with_conditions, key=lambda x: str(x.coin_spend))

        coin_spend_with_conditions = block_spends_with_conditions[1]

        assert coin_spend_with_conditions.coin_spend.coin == Coin(
            bytes.fromhex("e3b0c44298fc1c149afbf4c8996fb9240000000000000000000000000000000a"),
            bytes.fromhex("8488947a2213b2c2551fe019bbb708db86eab3dd5133eb57e801515e9e4ad82a"),
            1750000000000,
        )
        assert coin_spend_with_conditions.coin_spend.puzzle_reveal.to_program() == Program.fromhex(
            "ff02ffff01ff02ffff01ff02ffff03ff0bffff01ff02ffff03ffff09ff05ffff1dff0bffff1effff0bff0bffff02ff06ffff04ff02ffff04ff17ff8080808080808080ffff01ff02ff17ff2f80ffff01ff088080ff0180ffff01ff04ffff04ff04ffff04ff05ffff04ffff02ff06ffff04ff02ffff04ff17ff80808080ff80808080ffff02ff17ff2f808080ff0180ffff04ffff01ff32ff02ffff03ffff07ff0580ffff01ff0bffff0102ffff02ff06ffff04ff02ffff04ff09ff80808080ffff02ff06ffff04ff02ffff04ff0dff8080808080ffff01ff0bffff0101ff058080ff0180ff018080ffff04ffff01b0a499b52c7eba3465c3d74070a25d5ac5f5df25ed07d1c9c0c0509b00140da3e3bb60b584eaa30a1204ec0e5839f1252aff018080"
        )
        assert coin_spend_with_conditions.coin_spend.solution.to_program() == Program.fromhex(
            "ff80ffff01ffff33ffa063c767818f8b7cc8f3760ce34a09b7f34cd9ddf09d345c679b6897e7620c575cff8601977420dc0080ffff3cffa0a2366d6d8e1ce7496175528f5618a13da8401b02f2bac1eaae8f28aea9ee54798080ff8080"
        )
        assert coin_spend_with_conditions.conditions == [
            ConditionWithArgs(
                b"2",
                [
                    bytes.fromhex(
                        "a499b52c7eba3465c3d74070a25d5ac5f5df25ed07d1c9c0c0509b00140da3e3bb60b584eaa30a1204ec0e5839f1252a"
                    ),
                    bytes.fromhex("49b6f533000b967f049bb6e7b29d0b6f465ebccd5733bc75340f98dae782aa08"),
                ],
            ),
            ConditionWithArgs(
                b"3",
                [
                    bytes.fromhex("63c767818f8b7cc8f3760ce34a09b7f34cd9ddf09d345c679b6897e7620c575c"),
                    bytes.fromhex("01977420dc00"),
                ],
            ),
            ConditionWithArgs(
                b"<",
                [
                    bytes.fromhex("a2366d6d8e1ce7496175528f5618a13da8401b02f2bac1eaae8f28aea9ee5479"),
                ],
            ),
        ]

        coin_spend_with_conditions = block_spends_with_conditions[2]

        assert coin_spend_with_conditions.coin_spend.coin == Coin(
            bytes.fromhex("e3b0c44298fc1c149afbf4c8996fb9240000000000000000000000000000000b"),
            bytes.fromhex("8488947a2213b2c2551fe019bbb708db86eab3dd5133eb57e801515e9e4ad82a"),
            1750000000000,
        )
        assert coin_spend_with_conditions.coin_spend.puzzle_reveal.to_program() == Program.fromhex(
            "ff02ffff01ff02ffff01ff02ffff03ff0bffff01ff02ffff03ffff09ff05ffff1dff0bffff1effff0bff0bffff02ff06ffff04ff02ffff04ff17ff8080808080808080ffff01ff02ff17ff2f80ffff01ff088080ff0180ffff01ff04ffff04ff04ffff04ff05ffff04ffff02ff06ffff04ff02ffff04ff17ff80808080ff80808080ffff02ff17ff2f808080ff0180ffff04ffff01ff32ff02ffff03ffff07ff0580ffff01ff0bffff0102ffff02ff06ffff04ff02ffff04ff09ff80808080ffff02ff06ffff04ff02ffff04ff0dff8080808080ffff01ff0bffff0101ff058080ff0180ff018080ffff04ffff01b0a499b52c7eba3465c3d74070a25d5ac5f5df25ed07d1c9c0c0509b00140da3e3bb60b584eaa30a1204ec0e5839f1252aff018080"
        )
        assert coin_spend_with_conditions.coin_spend.solution.to_program() == Program.fromhex(
            "ff80ffff01ffff33ffa063c767818f8b7cc8f3760ce34a09b7f34cd9ddf09d345c679b6897e7620c575cff8601977420dc0080ffff3cffa04f6d4d12e97e83b2024fd0970e3b9e8a1c2e509625c15ff4145940c45b51974f8080ff8080"
        )
        assert coin_spend_with_conditions.conditions == [
            ConditionWithArgs(
                b"2",
                [
                    bytes.fromhex(
                        "a499b52c7eba3465c3d74070a25d5ac5f5df25ed07d1c9c0c0509b00140da3e3bb60b584eaa30a1204ec0e5839f1252a"
                    ),
                    bytes.fromhex("95df50b31bb746a37df6ab448f10436fb98bb659990c61ee6933a196f6a06465"),
                ],
            ),
            ConditionWithArgs(
                b"3",
                [
                    bytes.fromhex("63c767818f8b7cc8f3760ce34a09b7f34cd9ddf09d345c679b6897e7620c575c"),
                    bytes.fromhex("01977420dc00"),
                ],
            ),
            ConditionWithArgs(
                b"<",
                [
                    bytes.fromhex("4f6d4d12e97e83b2024fd0970e3b9e8a1c2e509625c15ff4145940c45b51974f"),
                ],
            ),
        ]

        coin_spend_with_conditions = block_spends_with_conditions[0]

        assert coin_spend_with_conditions.coin_spend.coin == Coin(
            bytes.fromhex("27ae41e4649b934ca495991b7852b8550000000000000000000000000000000b"),
            bytes.fromhex("8488947a2213b2c2551fe019bbb708db86eab3dd5133eb57e801515e9e4ad82a"),
            250000000000,
        )
        assert coin_spend_with_conditions.coin_spend.puzzle_reveal.to_program() == Program.fromhex(
            "ff02ffff01ff02ffff01ff02ffff03ff0bffff01ff02ffff03ffff09ff05ffff1dff0bffff1effff0bff0bffff02ff06ffff04ff02ffff04ff17ff8080808080808080ffff01ff02ff17ff2f80ffff01ff088080ff0180ffff01ff04ffff04ff04ffff04ff05ffff04ffff02ff06ffff04ff02ffff04ff17ff80808080ff80808080ffff02ff17ff2f808080ff0180ffff04ffff01ff32ff02ffff03ffff07ff0580ffff01ff0bffff0102ffff02ff06ffff04ff02ffff04ff09ff80808080ffff02ff06ffff04ff02ffff04ff0dff8080808080ffff01ff0bffff0101ff058080ff0180ff018080ffff04ffff01b0a499b52c7eba3465c3d74070a25d5ac5f5df25ed07d1c9c0c0509b00140da3e3bb60b584eaa30a1204ec0e5839f1252aff018080"
        )
        assert coin_spend_with_conditions.coin_spend.solution.to_program() == Program.fromhex(
            "ff80ffff01ffff33ffa063c767818f8b7cc8f3760ce34a09b7f34cd9ddf09d345c679b6897e7620c575cff853a3529440080ffff3cffa0617d9951551dc9e329fcab835f37fe4602c9ea57626cc2069228793f7007716f8080ff8080"
        )
        assert coin_spend_with_conditions.conditions == [
            ConditionWithArgs(
                b"2",
                [
                    bytes.fromhex(
                        "a499b52c7eba3465c3d74070a25d5ac5f5df25ed07d1c9c0c0509b00140da3e3bb60b584eaa30a1204ec0e5839f1252a"
                    ),
                    bytes.fromhex("f3dd65f1ca4b030a726182e0194174fe95ff7a66f54381cad3aab168b8e75ee7"),
                ],
            ),
            ConditionWithArgs(
                b"3",
                [
                    bytes.fromhex("63c767818f8b7cc8f3760ce34a09b7f34cd9ddf09d345c679b6897e7620c575c"),
                    bytes.fromhex("3a35294400"),
                ],
            ),
            ConditionWithArgs(
                b"<",
                [
                    bytes.fromhex("617d9951551dc9e329fcab835f37fe4602c9ea57626cc2069228793f7007716f"),
                ],
            ),
        ]

        memo = 32 * b"\f"

        for i in range(2):
            await full_node_api_1.farm_new_transaction_block(FarmNewBlockProtocol(ph_2))

            state = await client.get_blockchain_state()
            block = await client.get_block(state["peak"].header_hash)

            coin_to_spend = block.get_included_reward_coins()[0]

            spend_bundle = wallet.generate_signed_transaction(coin_to_spend.amount, ph_2, coin_to_spend, memo=memo)
            await client.push_tx(spend_bundle)

        await full_node_api_1.farm_new_transaction_block(FarmNewBlockProtocol(ph_2))

        coin_to_spend = (await client.get_coin_records_by_hint(memo))[0].coin

        # Spend the most recent coin so we can test including spent coins later
        spend_bundle = wallet.generate_signed_transaction(coin_to_spend.amount, ph_2, coin_to_spend, memo=memo)
        await client.push_tx(spend_bundle)

        await full_node_api_1.farm_new_transaction_block(FarmNewBlockProtocol(ph_2))

        coin_records = await client.get_coin_records_by_hint(memo)

        assert len(coin_records) == 3

        coin_records = await client.get_coin_records_by_hint(memo, include_spent_coins=False)

        assert len(coin_records) == 2

        state = await client.get_blockchain_state()

        # Get coin records by hint
        coin_records = await client.get_coin_records_by_hint(
            memo, start_height=state["peak"].height - 1, end_height=state["peak"].height
        )

        assert len(coin_records) == 1

        assert len(await client.get_connections()) == 0

        await client.open_connection(self_hostname, server_2._port)

        async def num_connections():
            return len(await client.get_connections())

        await time_out_assert(10, num_connections, 1)
        connections = await client.get_connections()
        assert NodeType(connections[0]["type"]) == NodeType.FULL_NODE.value
        assert len(await client.get_connections(NodeType.FULL_NODE)) == 1
        assert len(await client.get_connections(NodeType.FARMER)) == 0
        await client.close_connection(connections[0]["node_id"])
        await time_out_assert(10, num_connections, 0)

        blocks: list[FullBlock] = await client.get_blocks(0, 5)
        assert len(blocks) == 5

        await full_node_api_1.reorg_from_index_to_new_index(ReorgProtocol(2, 55, bytes([0x2] * 32), None))
        new_blocks_0: list[FullBlock] = await client.get_blocks(0, 5)
        assert len(new_blocks_0) == 7

        new_blocks: list[FullBlock] = await client.get_blocks(0, 5, exclude_reorged=True)
        assert len(new_blocks) == 5
        assert blocks[0].header_hash == new_blocks[0].header_hash
        assert blocks[1].header_hash == new_blocks[1].header_hash
        assert blocks[2].header_hash == new_blocks[2].header_hash
        assert blocks[3].header_hash != new_blocks[3].header_hash


@pytest.mark.anyio
async def test_signage_points(two_nodes_sim_and_wallets_services, empty_blockchain):
    nodes, _, bt = two_nodes_sim_and_wallets_services
    full_node_service_1, full_node_service_2 = nodes
    full_node_api_1 = full_node_service_1._api
    full_node_api_2 = full_node_service_2._api
    server_1 = full_node_api_1.full_node.server
    server_2 = full_node_api_2.full_node.server

    config = bt.config
    self_hostname = config["self_hostname"]

    peer = await connect_and_get_peer(server_1, server_2, self_hostname)

    async with FullNodeRpcClient.create_as_context(
        self_hostname,
        full_node_service_1.rpc_server.listen_port,
        full_node_service_1.root_path,
        full_node_service_1.config,
    ) as client:
        # Only provide one
        res = await client.get_recent_signage_point_or_eos(None, None)
        assert res is None
        res = await client.get_recent_signage_point_or_eos(std_hash(b"0"), std_hash(b"1"))
        assert res is None

        # Not found
        res = await client.get_recent_signage_point_or_eos(std_hash(b"0"), None)
        assert res is None
        res = await client.get_recent_signage_point_or_eos(None, std_hash(b"0"))
        assert res is None

        blocks = bt.get_consecutive_blocks(5)
        for block in blocks:
            await full_node_api_1.full_node.add_block(block)

        blocks = bt.get_consecutive_blocks(1, block_list_input=blocks, skip_slots=1, force_overflow=True)

        blockchain = full_node_api_1.full_node.blockchain
        second_blockchain = empty_blockchain

        for block in blocks:
            await _validate_and_add_block(second_blockchain, block)

        # Creates a signage point based on the last block
        peak_2 = second_blockchain.get_peak()
        sp: SignagePoint = get_signage_point(
            bt.constants,
            blockchain,
            peak_2,
            peak_2.ip_sub_slot_total_iters(bt.constants),
            uint8(4),
            [],
            peak_2.sub_slot_iters,
        )

        # Don't have SP yet
        res = await client.get_recent_signage_point_or_eos(sp.cc_vdf.output.get_hash(), None)
        assert res is None

        # Add the last block
        await full_node_api_1.full_node.add_block(blocks[-1])
        await full_node_api_1.respond_signage_point(
            full_node_protocol.RespondSignagePoint(uint8(4), sp.cc_vdf, sp.cc_proof, sp.rc_vdf, sp.rc_proof), peer
        )

        assert full_node_api_1.full_node.full_node_store.get_signage_point(sp.cc_vdf.output.get_hash()) is not None

        # Properly fetch a signage point
        res = await client.get_recent_signage_point_or_eos(sp.cc_vdf.output.get_hash(), None)

        assert res is not None
        assert "eos" not in res
        assert res["signage_point"] == sp
        assert not res["reverted"]

        blocks = bt.get_consecutive_blocks(1, block_list_input=blocks, skip_slots=1)
        selected_eos = blocks[-1].finished_sub_slots[0]

        # Don't have EOS yet
        res = await client.get_recent_signage_point_or_eos(None, selected_eos.challenge_chain.get_hash())
        assert res is None

        # Properly fetch an EOS
        for eos in blocks[-1].finished_sub_slots:
            await full_node_api_1.full_node.add_end_of_sub_slot(eos, peer)

        res = await client.get_recent_signage_point_or_eos(None, selected_eos.challenge_chain.get_hash())
        assert res is not None
        assert "signage_point" not in res
        assert res["eos"] == selected_eos
        assert not res["reverted"]

        # Do another one but without sending the slot
        await full_node_api_1.full_node.add_block(blocks[-1])
        blocks = bt.get_consecutive_blocks(1, block_list_input=blocks, skip_slots=2)
        selected_eos = blocks[-1].finished_sub_slots[-1]
        await full_node_api_1.full_node.add_block(blocks[-1])

        res = await client.get_recent_signage_point_or_eos(None, selected_eos.challenge_chain.get_hash())
        assert res is not None
        assert "signage_point" not in res
        assert res["eos"] == selected_eos
        assert not res["reverted"]

        # Perform a reorg
        blocks = bt.get_consecutive_blocks(12, seed=b"1234")
        await add_blocks_in_batches(blocks, full_node_api_1.full_node)

        # Signage point is no longer in the blockchain
        res = await client.get_recent_signage_point_or_eos(sp.cc_vdf.output.get_hash(), None)
        assert res["reverted"]
        assert res["signage_point"] == sp
        assert "eos" not in res

        # EOS is no longer in the blockchain
        res = await client.get_recent_signage_point_or_eos(None, selected_eos.challenge_chain.get_hash())
        assert res is not None
        assert "signage_point" not in res
        assert res["eos"] == selected_eos
        assert res["reverted"]


@pytest.mark.anyio
async def test_get_network_info(one_wallet_and_one_simulator_services, self_hostname):
    nodes, _, _bt = one_wallet_and_one_simulator_services
    (full_node_service_1,) = nodes

    async with FullNodeRpcClient.create_as_context(
        self_hostname,
        full_node_service_1.rpc_server.listen_port,
        full_node_service_1.root_path,
        full_node_service_1.config,
    ) as client:
        await validate_get_routes(client, full_node_service_1.rpc_server.rpc_api)
        network_info = await client.fetch("get_network_info", {})
        assert network_info == {
            "network_name": "testnet0",
            "network_prefix": "txck",
            "genesis_challenge": "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
            "success": True,
        }


@pytest.mark.anyio
async def test_get_version(one_wallet_and_one_simulator_services, self_hostname):
    nodes, _, _bt = one_wallet_and_one_simulator_services
    (full_node_service_1,) = nodes
    async with FullNodeRpcClient.create_as_context(
        self_hostname,
        full_node_service_1.rpc_server.listen_port,
        full_node_service_1.root_path,
        full_node_service_1.config,
    ) as client:
        version = await client.fetch("get_version", {})
        assert version == {
            "success": True,
            "version": __version__,
        }


@pytest.mark.anyio
async def test_get_blockchain_state(one_wallet_and_one_simulator_services, self_hostname):
    num_blocks = 5
    nodes, _, bt = one_wallet_and_one_simulator_services
    (full_node_service_1,) = nodes
    full_node_api_1 = full_node_service_1._api

    try:
        client = await FullNodeRpcClient.create(
            self_hostname,
            full_node_service_1.rpc_server.listen_port,
            full_node_service_1.root_path,
            full_node_service_1.config,
        )
        await validate_get_routes(client, full_node_service_1.rpc_server.rpc_api)
        state = await client.get_blockchain_state()
        assert state["peak"] is None
        assert not state["sync"]["sync_mode"]
        assert state["difficulty"] > 0
        assert state["sub_slot_iters"] > 0
        assert state["space"] == 0
        assert state["average_block_time"] is None

        blocks: list[FullBlock] = bt.get_consecutive_blocks(num_blocks)
        blocks = bt.get_consecutive_blocks(num_blocks, block_list_input=blocks, guarantee_transaction_block=True)

        for block in blocks:
            unf = UnfinishedBlock(
                block.finished_sub_slots,
                block.reward_chain_block.get_unfinished(),
                block.challenge_chain_sp_proof,
                block.reward_chain_sp_proof,
                block.foliage,
                block.foliage_transaction_block,
                block.transactions_info,
                block.transactions_generator,
                [],
            )
            await full_node_api_1.full_node.add_unfinished_block(unf, None)
            await full_node_api_1.full_node.add_block(block, None)

        state = await client.get_blockchain_state()

        assert state["space"] > 0
        assert state["average_block_time"] > 0

        block_records: list[BlockRecord] = [
            await full_node_api_1.full_node.blockchain.get_block_record_from_db(rec.header_hash) for rec in blocks
        ]
        first_non_transaction_block_index = -1
        for i, b in enumerate(block_records):
            if not b.is_transaction_block:
                first_non_transaction_block_index = i
                break
        # Genesis block(height=0) must be a transaction block
        # so first_non_transaction_block_index != 0
        assert first_non_transaction_block_index > 0

        transaction_blocks: list[BlockRecord] = [b for b in block_records if b.is_transaction_block]
        non_transaction_block: list[BlockRecord] = [b for b in block_records if not b.is_transaction_block]
        assert len(transaction_blocks) > 0
        assert len(non_transaction_block) > 0
        assert transaction_blocks[0] == await get_nearest_transaction_block(
            full_node_api_1.full_node.blockchain, transaction_blocks[0]
        )

        nearest_transaction_block = block_records[first_non_transaction_block_index - 1]
        expected_nearest_transaction_block = await get_nearest_transaction_block(
            full_node_api_1.full_node.blockchain, block_records[first_non_transaction_block_index]
        )
        assert expected_nearest_transaction_block == nearest_transaction_block
        # When supplying genesis block, there are no older blocks so `None` should be returned
        assert await get_average_block_time(full_node_api_1.full_node.blockchain, block_records[0], 4608) is None
        assert await get_average_block_time(full_node_api_1.full_node.blockchain, block_records[-1], 4608) is not None
        # Test that get_aggsig_additional_data() returns correctly
        assert (
            full_node_api_1.full_node.constants.AGG_SIG_ME_ADDITIONAL_DATA == await client.get_aggsig_additional_data()
        )

    finally:
        # Checks that the RPC manages to stop the node
        client.close()
        await client.await_closed()


@pytest.mark.anyio
async def test_coin_name_not_in_request(one_node, self_hostname):
    [full_node_service], _, _ = one_node

    async with FullNodeRpcClient.create_as_context(
        self_hostname,
        full_node_service.rpc_server.listen_port,
        full_node_service.root_path,
        full_node_service.config,
    ) as client:
        with pytest.raises(ValueError, match="No coin_name in request"):
            await client.fetch("get_mempool_items_by_coin_name", {})


@pytest.mark.anyio
async def test_coin_name_not_found_in_mempool(one_node, self_hostname):
    [full_node_service], _, _ = one_node

    async with FullNodeRpcClient.create_as_context(
        self_hostname,
        full_node_service.rpc_server.listen_port,
        full_node_service.root_path,
        full_node_service.config,
    ) as client:
        empty_coin_name = bytes32.zeros
        mempool_item = await client.get_mempool_items_by_coin_name(empty_coin_name)
        assert mempool_item["success"]
        assert "mempool_items" in mempool_item
        assert len(mempool_item["mempool_items"]) == 0


@pytest.mark.anyio
async def test_coin_name_found_in_mempool(one_node, self_hostname):
    [full_node_service], _, bt = one_node
    full_node_api = full_node_service._api

    async with FullNodeRpcClient.create_as_context(
        self_hostname,
        full_node_service.rpc_server.listen_port,
        full_node_service.root_path,
        full_node_service.config,
    ) as client:
        blocks = bt.get_consecutive_blocks(2)
        blocks = bt.get_consecutive_blocks(2, block_list_input=blocks, guarantee_transaction_block=True)

        for block in blocks:
            unf = UnfinishedBlock(
                block.finished_sub_slots,
                block.reward_chain_block.get_unfinished(),
                block.challenge_chain_sp_proof,
                block.reward_chain_sp_proof,
                block.foliage,
                block.foliage_transaction_block,
                block.transactions_info,
                block.transactions_generator,
                [],
            )
            await full_node_api.full_node.add_unfinished_block(unf, None)
            await full_node_api.full_node.add_block(block, None)

        wallet = WalletTool(full_node_api.full_node.constants)
        wallet_receiver = WalletTool(full_node_api.full_node.constants, AugSchemeMPL.key_gen(std_hash(b"123123")))
        ph = wallet.get_new_puzzlehash()
        ph_receiver = wallet_receiver.get_new_puzzlehash()

        blocks = bt.get_consecutive_blocks(
            2,
            block_list_input=blocks,
            guarantee_transaction_block=True,
            farmer_reward_puzzle_hash=ph,
            pool_reward_puzzle_hash=ph,
        )
        for block in blocks[-2:]:
            await full_node_api.full_node.add_block(block)

        # empty mempool
        assert len(await client.get_all_mempool_items()) == 0

        coin_to_spend = blocks[-1].get_included_reward_coins()[0]
        spend_bundle = wallet.generate_signed_transaction(coin_to_spend.amount, ph_receiver, coin_to_spend)
        await client.push_tx(spend_bundle)

        # mempool with one item
        assert len(await client.get_all_mempool_items()) == 1

        mempool_item = await client.get_mempool_items_by_coin_name(coin_to_spend.name())

        # found coin in coin spends
        assert mempool_item["success"]
        assert "mempool_items" in mempool_item
        assert len(mempool_item["mempool_items"]) > 0
        for item in mempool_item["mempool_items"]:
            removals = [Coin.from_json_dict(coin) for coin in item["removals"]]
            assert coin_to_spend.name() in [coin.name() for coin in removals]



from __future__ import annotations

import json
from pathlib import Path
from typing import Any, Optional, Union

from chik.rpc.full_node_rpc_client import FullNodeRpcClient
from chik.types.blockchain_format.sized_bytes import bytes32


async def print_blockchain_state(node_client: FullNodeRpcClient, config: dict[str, Any]) -> bool:
    import time

    from chik.cmds.cmds_util import format_bytes
    from chik.consensus.block_record import BlockRecord
    from chik.util.ints import uint64

    blockchain_state = await node_client.get_blockchain_state()
    if blockchain_state is None:
        print("There is no blockchain found yet. Try again shortly")
        return True
    peak: Optional[BlockRecord] = blockchain_state["peak"]
    node_id = blockchain_state["node_id"]
    difficulty = blockchain_state["difficulty"]
    sub_slot_iters = blockchain_state["sub_slot_iters"]
    synced = blockchain_state["sync"]["synced"]
    sync_mode = blockchain_state["sync"]["sync_mode"]
    num_blocks: int = 10
    network_name = config["selected_network"]
    genesis_challenge = config["farmer"]["network_overrides"]["constants"][network_name]["GENESIS_CHALLENGE"]
    full_node_port = config["full_node"]["port"]
    full_node_rpc_port = config["full_node"]["rpc_port"]

    print(f"Network: {network_name}    Port: {full_node_port}   RPC Port: {full_node_rpc_port}")
    print(f"Node ID: {node_id}")
    print(f"Genesis Challenge: {genesis_challenge}")

    if synced:
        print("Current Blockchain Status: Full Node Synced")
        print("\nPeak: Hash:", bytes32(peak.header_hash) if peak is not None else "")
    elif peak is not None and sync_mode:
        sync_max_block = blockchain_state["sync"]["sync_tip_height"]
        sync_current_block = blockchain_state["sync"]["sync_progress_height"]
        print(
            f"Current Blockchain Status: Syncing {sync_current_block}/{sync_max_block} "
            f"({sync_max_block - sync_current_block} behind). "
            f"({sync_current_block * 100.0 / sync_max_block:2.2f}% synced)"
        )
        print("Peak: Hash:", bytes32(peak.header_hash) if peak is not None else "")
    elif peak is not None:
        print(f"Current Blockchain Status: Not Synced. Peak height: {peak.height}")
    else:
        print("\nSearching for an initial chain\n")
        print("You may be able to expedite with 'chik peer full_node -a host:port' using a known node.\n")

    if peak is not None:
        if peak.is_transaction_block:
            peak_time = peak.timestamp
        else:
            peak_hash = bytes32(peak.header_hash)
            curr = await node_client.get_block_record(peak_hash)
            while curr is not None and not curr.is_transaction_block:
                curr = await node_client.get_block_record(curr.prev_hash)
            if curr is not None:
                peak_time = curr.timestamp
            else:
                peak_time = uint64(0)
        peak_time_struct = time.struct_time(time.localtime(peak_time))

        print(
            "      Time:",
            f"{time.strftime('%a %b %d %Y %T %Z', peak_time_struct)}",
            f"                 Height: {peak.height:>10}\n",
        )

        print("Estimated network space: ", end="")
        print(format_bytes(blockchain_state["space"]))
        print(f"Current difficulty: {difficulty}")
        print(f"Current VDF sub_slot_iters: {sub_slot_iters}")
        print("\n  Height: |   Hash:")

        added_blocks: list[BlockRecord] = []
        curr = await node_client.get_block_record(peak.header_hash)
        while curr is not None and len(added_blocks) < num_blocks and curr.height > 0:
            added_blocks.append(curr)
            curr = await node_client.get_block_record(curr.prev_hash)

        for b in added_blocks:
            print(f"{b.height:>9} | {bytes32(b.header_hash)}")
    else:
        print("Blockchain has no blocks yet")
    return False


async def print_block_from_hash(
    node_client: FullNodeRpcClient, config: dict[str, Any], block_by_header_hash: str
) -> None:
    import time

    from chik.consensus.block_record import BlockRecord
    from chik.types.blockchain_format.sized_bytes import bytes32
    from chik.types.full_block import FullBlock
    from chik.util.bech32m import encode_puzzle_hash

    block: Optional[BlockRecord] = await node_client.get_block_record(bytes32.from_hexstr(block_by_header_hash))
    full_block: Optional[FullBlock] = await node_client.get_block(bytes32.from_hexstr(block_by_header_hash))
    # Would like to have a verbose flag for this
    if block is not None:
        assert full_block is not None
        prev_b = await node_client.get_block_record(block.prev_hash)
        if prev_b is not None:
            difficulty = block.weight - prev_b.weight
        else:
            difficulty = block.weight
        if block.is_transaction_block:
            assert full_block.transactions_info is not None
            block_time = time.struct_time(
                time.localtime(
                    full_block.foliage_transaction_block.timestamp if full_block.foliage_transaction_block else None
                )
            )
            block_time_string = time.strftime("%a %b %d %Y %T %Z", block_time)
            cost = str(full_block.transactions_info.cost)
            tx_filter_hash: Union[str, bytes32] = "Not a transaction block"
            if full_block.foliage_transaction_block:
                tx_filter_hash = bytes32(full_block.foliage_transaction_block.filter_hash)
            fees: Any = block.fees
        else:
            block_time_string = "Not a transaction block"
            cost = "Not a transaction block"
            tx_filter_hash = "Not a transaction block"
            fees = "Not a transaction block"
        address_prefix = config["network_overrides"]["config"][config["selected_network"]]["address_prefix"]
        farmer_address = encode_puzzle_hash(block.farmer_puzzle_hash, address_prefix)
        pool_address = encode_puzzle_hash(block.pool_puzzle_hash, address_prefix)
        pool_pk = (
            full_block.reward_chain_block.proof_of_space.pool_public_key
            if full_block.reward_chain_block.proof_of_space.pool_public_key is not None
            else "Pay to pool puzzle hash"
        )
        print(
            f"Block Height           {block.height}\n"
            f"Header Hash            0x{block.header_hash.hex()}\n"
            f"Timestamp              {block_time_string}\n"
            f"Weight                 {block.weight}\n"
            f"Previous Block         0x{block.prev_hash.hex()}\n"
            f"Difficulty             {difficulty}\n"
            f"Sub-slot iters         {block.sub_slot_iters}\n"
            f"Cost                   {cost}\n"
            f"Total VDF Iterations   {block.total_iters}\n"
            f"Is a Transaction Block?{block.is_transaction_block}\n"
            f"Deficit                {block.deficit}\n"
            f"PoSpace 'k' Size       {full_block.reward_chain_block.proof_of_space.size}\n"
            f"Plot Public Key        0x{full_block.reward_chain_block.proof_of_space.plot_public_key}\n"
            f"Pool Public Key        {pool_pk}\n"
            f"Tx Filter Hash         {tx_filter_hash}\n"
            f"Farmer Address         {farmer_address}\n"
            f"Pool Address           {pool_address}\n"
            f"Fees Amount            {fees}\n"
        )
    else:
        print("Block with header hash", block_by_header_hash, "not found")


async def print_fee_info(node_client: FullNodeRpcClient) -> None:
    target_times = [60, 120, 300]
    target_times_names = ["1  minute", "2 minutes", "5 minutes"]
    res = await node_client.get_fee_estimate(target_times=target_times, cost=1)
    print(json.dumps(res))
    print("\n")
    print(f"  Mempool max cost: {res['mempool_max_size']:>12} KLVM cost")
    print(f"      Mempool cost: {res['mempool_size']:>12} KLVM cost")
    print(f"     Mempool count: {res['num_spends']:>12} spends")
    print(f"   Fees in Mempool: {res['mempool_fees']:>12} mojos")
    print()

    print("Stats for last transaction block:")
    print(f"      Block height: {res['last_tx_block_height']:>12}")
    print(f"        Block fees: {res['fees_last_block']:>12} mojos")
    print(f"        Block cost: {res['last_block_cost']:>12} KLVM cost")
    print(f"          Fee rate: {res['fee_rate_last_block']:>12.5} mojos per KLVM cost")

    print("\nFee Rate Estimates:")
    max_name_len = max(len(name) for name in target_times_names)
    for n, e in zip(target_times_names, res["estimates"]):
        print(f"    {n:>{max_name_len}}: {e:.3f} mojo per KLVM cost")
    print("")


async def show_async(
    rpc_port: Optional[int],
    root_path: Path,
    print_fee_info_flag: bool,
    print_state: bool,
    block_header_hash_by_height: Optional[int],
    block_by_header_hash: str,
) -> None:
    from chik.cmds.cmds_util import get_any_service_client

    async with get_any_service_client(FullNodeRpcClient, root_path, rpc_port) as (node_client, config):
        # Check State
        if print_state:
            if await print_blockchain_state(node_client, config) is True:
                return None  # if no blockchain is found
        if print_fee_info_flag:
            await print_fee_info(node_client)
        # Get Block Information
        if block_header_hash_by_height is not None:
            block_header = await node_client.get_block_record_by_height(block_header_hash_by_height)
            if block_header is not None:
                print(f"Header hash of block {block_header_hash_by_height}: {block_header.header_hash.hex()}")
            else:
                print("Block height", block_header_hash_by_height, "not found")
        if block_by_header_hash != "":
            await print_block_from_hash(node_client, config, block_by_header_hash)



from __future__ import annotations

import logging
from typing import Union

from chik.consensus.block_record import BlockRecord
from chik.consensus.blockchain_interface import BlockRecordsProtocol
from chik.consensus.constants import ConsensusConstants
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.full_block import FullBlock
from chik.types.header_block import HeaderBlock
from chik.types.unfinished_block import UnfinishedBlock
from chik.types.unfinished_header_block import UnfinishedHeaderBlock
from chik.util.ints import uint64

log = logging.getLogger(__name__)


def final_eos_is_already_included(
    header_block: Union[UnfinishedHeaderBlock, UnfinishedBlock, HeaderBlock, FullBlock],
    blocks: BlockRecordsProtocol,
    sub_slot_iters: uint64,
) -> bool:
    """
    Args:
        header_block: An overflow block, with potentially missing information about the new sub slot
        blocks: all blocks that have been included before header_block
        sub_slot_iters: sub_slot_iters at the header_block

    Returns: True iff the missing sub slot was already included in a previous block. Returns False if the sub
    slot was not included yet, and therefore it is the responsibility of this block to include it

    """
    if len(header_block.finished_sub_slots) > 0:
        # We already have an included empty sub slot, which means the prev block is 2 sub slots behind.
        return False
    curr: BlockRecord = blocks.block_record(header_block.prev_header_hash)

    # We also check if curr is close to header_block, which means it's in the same sub slot
    seen_overflow_block = curr.overflow and (header_block.total_iters - curr.total_iters < sub_slot_iters // 2)
    while not curr.first_in_sub_slot and not curr.height == 0:
        if curr.overflow and header_block.total_iters - curr.total_iters < sub_slot_iters // 2:
            seen_overflow_block = True
        curr = blocks.block_record(curr.prev_hash)

    if curr.first_in_sub_slot and seen_overflow_block:
        # We have seen another overflow block in this slot (same as header_block), therefore there are no
        # missing sub slots
        return True

    # We have not seen any overflow blocks, therefore header_block will have to include the missing sub slot in
    # the future
    return False


def get_block_challenge(
    constants: ConsensusConstants,
    header_block: Union[UnfinishedHeaderBlock, UnfinishedBlock, HeaderBlock, FullBlock],
    blocks: BlockRecordsProtocol,
    genesis_block: bool,
    overflow: bool,
    skip_overflow_last_ss_validation: bool,
) -> bytes32:
    if len(header_block.finished_sub_slots) > 0:
        if overflow:
            # New sub-slot with overflow block
            if skip_overflow_last_ss_validation:
                # In this case, we are missing the final sub-slot bundle (it's not finished yet), however
                # There is a whole empty slot before this block is infused
                challenge: bytes32 = header_block.finished_sub_slots[-1].challenge_chain.get_hash()
            else:
                challenge = header_block.finished_sub_slots[
                    -1
                ].challenge_chain.challenge_chain_end_of_slot_vdf.challenge
        else:
            # No overflow, new slot with a new challenge
            challenge = header_block.finished_sub_slots[-1].challenge_chain.get_hash()
    else:
        if genesis_block:
            challenge = constants.GENESIS_CHALLENGE
        else:
            if overflow:
                if skip_overflow_last_ss_validation:
                    # Overflow infusion without the new slot, so get the last challenge
                    challenges_to_look_for = 1
                else:
                    # Overflow infusion, so get the second to last challenge. skip_overflow_last_ss_validation is False,
                    # Which means no sub slots are omitted
                    challenges_to_look_for = 2
            else:
                challenges_to_look_for = 1
            reversed_challenge_hashes: list[bytes32] = []
            curr: BlockRecord = blocks.block_record(header_block.prev_header_hash)
            while len(reversed_challenge_hashes) < challenges_to_look_for:
                if curr.first_in_sub_slot:
                    assert curr.finished_challenge_slot_hashes is not None
                    reversed_challenge_hashes += reversed(curr.finished_challenge_slot_hashes)
                    if len(reversed_challenge_hashes) >= challenges_to_look_for:
                        break
                if curr.height == 0:
                    assert curr.finished_challenge_slot_hashes is not None
                    assert len(curr.finished_challenge_slot_hashes) > 0
                    break
                curr = blocks.block_record(curr.prev_hash)
            challenge = reversed_challenge_hashes[challenges_to_look_for - 1]
    return challenge



from __future__ import annotations

import pytest

from chik.util.setproctitle import setproctitle

pytestmark = pytest.mark.skip(
    reason="this test ends up hanging frequently and needs to be rewritten with a subprocess and a title check",
)


def test_does_not_crash() -> None:
    setproctitle("chik test title")



from __future__ import annotations

import logging
from collections.abc import AsyncGenerator
from dataclasses import replace
from typing import Any

import pytest

from chik.daemon.keychain_proxy import KeychainProxy, connect_to_keychain_and_validate
from chik.simulator.block_tools import BlockTools
from chik.simulator.setup_services import setup_daemon
from chik.util.errors import KeychainIsEmpty, KeychainKeyNotFound
from chik.util.keychain import KeyData

TEST_KEY_1 = KeyData.generate(label="🚽🍯")
TEST_KEY_2 = KeyData.generate(label="👨‍✈️🥦")
TEST_KEY_3 = KeyData.generate(label="☕️🍬")


@pytest.fixture(scope="function", params=[True, False])
async def keychain_proxy(get_b_tools: BlockTools, request: Any) -> AsyncGenerator[KeychainProxy, None]:
    async with setup_daemon(btools=get_b_tools) as daemon:
        log = logging.getLogger("keychain_proxy_fixture")
        keychain_proxy = await connect_to_keychain_and_validate(daemon.root_path, log)
        assert keychain_proxy is not None
        if request.param:
            keychain_proxy.keychain = daemon.keychain_server._default_keychain
        yield keychain_proxy
        await keychain_proxy.close()


@pytest.fixture(scope="function")
async def keychain_proxy_with_keys(keychain_proxy: KeychainProxy) -> KeychainProxy:
    await keychain_proxy.add_key(TEST_KEY_1.mnemonic_str(), TEST_KEY_1.label)
    await keychain_proxy.add_key(TEST_KEY_2.mnemonic_str(), TEST_KEY_2.label)
    return keychain_proxy


@pytest.mark.anyio
async def test_add_private_key(keychain_proxy: KeychainProxy) -> None:
    keychain = keychain_proxy
    await keychain.add_key(TEST_KEY_3.mnemonic_str(), TEST_KEY_3.label)
    key = await keychain.get_key(TEST_KEY_3.fingerprint, include_secrets=True)
    assert key == TEST_KEY_3


@pytest.mark.anyio
async def test_add_public_key(keychain_proxy: KeychainProxy) -> None:
    keychain = keychain_proxy
    await keychain.add_key(bytes(TEST_KEY_3.public_key).hex(), TEST_KEY_3.label, private=False)
    with pytest.raises(Exception, match="already exists"):
        await keychain.add_key(bytes(TEST_KEY_3.public_key).hex(), "", private=False)
    key = await keychain.get_key(TEST_KEY_3.fingerprint, include_secrets=False)
    assert key is not None
    assert key.public_key == TEST_KEY_3.public_key
    assert key.secrets is None

    pk = await keychain.get_key_for_fingerprint(TEST_KEY_3.fingerprint, private=False)
    assert pk is not None
    assert pk == TEST_KEY_3.public_key

    pk = await keychain.get_key_for_fingerprint(None, private=False)
    assert pk is not None
    assert pk == TEST_KEY_3.public_key

    with pytest.raises(KeychainKeyNotFound):
        pk = await keychain.get_key_for_fingerprint(1234567890, private=False)


@pytest.mark.parametrize("include_secrets", [True, False])
@pytest.mark.anyio
async def test_get_key(keychain_proxy_with_keys: KeychainProxy, include_secrets: bool) -> None:
    keychain = keychain_proxy_with_keys
    key = await keychain.get_key(TEST_KEY_1.fingerprint, include_secrets=include_secrets)
    expected_key = TEST_KEY_1 if include_secrets else replace(TEST_KEY_1, secrets=None)
    assert key == expected_key


@pytest.mark.anyio
async def test_get_key_for_fingerprint(keychain_proxy: KeychainProxy) -> None:
    keychain = keychain_proxy
    with pytest.raises(KeychainIsEmpty):
        await keychain.get_key_for_fingerprint(None, private=False)
    await keychain_proxy.add_key(TEST_KEY_1.mnemonic_str(), TEST_KEY_1.label)
    assert await keychain.get_key_for_fingerprint(TEST_KEY_1.fingerprint, private=False) == TEST_KEY_1.public_key
    assert await keychain.get_key_for_fingerprint(None, private=False) == TEST_KEY_1.public_key
    with pytest.raises(KeychainKeyNotFound):
        await keychain.get_key_for_fingerprint(1234567890, private=False)


@pytest.mark.parametrize("include_secrets", [True, False])
@pytest.mark.anyio
async def test_get_keys(keychain_proxy_with_keys: KeychainProxy, include_secrets: bool) -> None:
    keychain = keychain_proxy_with_keys
    keys = await keychain.get_keys(include_secrets=include_secrets)
    if include_secrets:
        expected_keys = [TEST_KEY_1, TEST_KEY_2]
    else:
        expected_keys = [replace(TEST_KEY_1, secrets=None), replace(TEST_KEY_2, secrets=None)]
    assert keys == expected_keys



from __future__ import annotations

import random
from dataclasses import dataclass, field
from typing import Optional

import pytest
from klvm_tools import binutils

from chik._tests.util.db_connection import DBConnection
from chik.types.blockchain_format.coin import Coin
from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.serialized_program import SerializedProgram
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.coin_spend import CoinSpend, compute_additions, make_spend
from chik.util.ints import uint32, uint64
from chik.wallet.wallet_pool_store import WalletPoolStore


def make_child_solution(
    coin_spend: Optional[CoinSpend], new_coin: Optional[Coin], seeded_random: random.Random
) -> CoinSpend:
    new_puzzle_hash: bytes32 = bytes32.random(seeded_random)
    solution = "()"
    puzzle = f"(q . ((51 0x{new_puzzle_hash.hex()} 1)))"
    puzzle_prog = Program.to(binutils.assemble(puzzle))
    solution_prog = Program.to(binutils.assemble(solution))
    if new_coin is None:
        assert coin_spend is not None
        new_coin = compute_additions(coin_spend)[0]
    sol: CoinSpend = make_spend(
        new_coin,
        SerializedProgram.from_program(puzzle_prog),
        SerializedProgram.from_program(solution_prog),
    )
    return sol


async def assert_db_spends(store: WalletPoolStore, wallet_id: int, spends: list[CoinSpend]) -> None:
    db_spends = await store.get_spends_for_wallet(wallet_id)
    assert len(db_spends) == len(spends)
    for spend, (_, db_spend) in zip(spends, db_spends):
        assert spend == db_spend


@dataclass
class DummySpends:
    seeded_random: random.Random
    spends_per_wallet: dict[int, list[CoinSpend]] = field(default_factory=dict)

    def generate(self, wallet_id: int, count: int) -> None:
        current = self.spends_per_wallet.setdefault(wallet_id, [])
        for _ in range(count):
            coin = None
            last_spend = None if len(current) == 0 else current[-1]
            if last_spend is None:
                coin = Coin(bytes32.random(self.seeded_random), bytes32.random(self.seeded_random), uint64(12312))
            current.append(make_child_solution(coin_spend=last_spend, new_coin=coin, seeded_random=self.seeded_random))


class TestWalletPoolStore:
    @pytest.mark.anyio
    async def test_store(self, seeded_random: random.Random):
        async with DBConnection(1) as db_wrapper:
            store = await WalletPoolStore.create(db_wrapper)

            try:
                async with db_wrapper.writer():
                    coin_0 = Coin(bytes32.random(seeded_random), bytes32.random(seeded_random), uint64(12312))
                    coin_0_alt = Coin(bytes32.random(seeded_random), bytes32.random(seeded_random), uint64(12312))
                    solution_0: CoinSpend = make_child_solution(
                        coin_spend=None, new_coin=coin_0, seeded_random=seeded_random
                    )
                    solution_0_alt: CoinSpend = make_child_solution(
                        coin_spend=None, new_coin=coin_0_alt, seeded_random=seeded_random
                    )
                    solution_1: CoinSpend = make_child_solution(
                        coin_spend=solution_0, new_coin=None, seeded_random=seeded_random
                    )

                    assert await store.get_spends_for_wallet(0) == []
                    assert await store.get_spends_for_wallet(1) == []

                    await store.add_spend(1, solution_1, 100)
                    assert await store.get_spends_for_wallet(1) == [(100, solution_1)]

                    # Idempotent
                    await store.add_spend(1, solution_1, 100)
                    assert await store.get_spends_for_wallet(1) == [(100, solution_1)]

                    with pytest.raises(ValueError):
                        await store.add_spend(1, solution_1, 101)

                    # Rebuild cache, no longer present
                    raise RuntimeError("abandon transaction")
            except Exception:
                pass

            assert await store.get_spends_for_wallet(1) == []

            await store.add_spend(1, solution_1, 100)
            assert await store.get_spends_for_wallet(1) == [(100, solution_1)]

            solution_1_alt: CoinSpend = make_child_solution(solution_0_alt, new_coin=None, seeded_random=seeded_random)

            with pytest.raises(ValueError):
                await store.add_spend(1, solution_1_alt, 100)

            assert await store.get_spends_for_wallet(1) == [(100, solution_1)]

            solution_2: CoinSpend = make_child_solution(solution_1, new_coin=None, seeded_random=seeded_random)
            await store.add_spend(1, solution_2, 100)
            solution_3: CoinSpend = make_child_solution(solution_2, new_coin=None, seeded_random=seeded_random)
            await store.add_spend(1, solution_3, 100)
            solution_4: CoinSpend = make_child_solution(solution_3, new_coin=None, seeded_random=seeded_random)

            with pytest.raises(ValueError):
                await store.add_spend(1, solution_4, 99)

            await store.add_spend(1, solution_4, 101)
            await store.rollback(101, 1)
            assert await store.get_spends_for_wallet(1) == [
                (100, solution_1),
                (100, solution_2),
                (100, solution_3),
                (101, solution_4),
            ]
            await store.rollback(100, 1)
            assert await store.get_spends_for_wallet(1) == [
                (100, solution_1),
                (100, solution_2),
                (100, solution_3),
            ]
            with pytest.raises(ValueError):
                await store.add_spend(1, solution_1, 105)

            await store.add_spend(1, solution_4, 105)
            solution_5: CoinSpend = make_child_solution(solution_4, new_coin=None, seeded_random=seeded_random)
            await store.add_spend(1, solution_5, 105)
            await store.rollback(99, 1)
            assert await store.get_spends_for_wallet(1) == []


@pytest.mark.anyio
async def test_delete_wallet(seeded_random: random.Random) -> None:
    dummy_spends = DummySpends(seeded_random=seeded_random)
    for i in range(5):
        dummy_spends.generate(i, i * 5)
    async with DBConnection(1) as db_wrapper:
        store = await WalletPoolStore.create(db_wrapper)
        # Add the spends per wallet and verify them
        for wallet_id, spends in dummy_spends.spends_per_wallet.items():
            for i, spend in enumerate(spends):
                await store.add_spend(wallet_id, spend, uint32(i + wallet_id))
            await assert_db_spends(store, wallet_id, spends)
        # Remove one wallet after the other and verify before and after each
        for wallet_id, spends in dummy_spends.spends_per_wallet.items():
            # Assert the existence again here to make sure the previous removals did not affect other wallet_ids
            await assert_db_spends(store, wallet_id, spends)
            await store.delete_wallet(wallet_id)
            await assert_db_spends(store, wallet_id, [])



from __future__ import annotations

from logging import Logger
from pathlib import Path

import pytest

from chik.full_node.full_node_api import FullNodeAPI
from chik.server.node_discovery import FullNodeDiscovery
from chik.server.server import ChikServer
from chik.simulator.block_tools import BlockTools
from chik.util.default_root import SIMULATOR_ROOT_PATH


@pytest.mark.anyio
async def test_enable_private_networks(
    two_nodes: tuple[FullNodeAPI, FullNodeAPI, ChikServer, ChikServer, BlockTools],
) -> None:
    chik_server = two_nodes[2]

    # Missing `enable_private_networks` config entry in introducer_peer should default to False for back compat
    discovery0 = FullNodeDiscovery(
        chik_server,
        0,
        SIMULATOR_ROOT_PATH / Path(chik_server.config["peers_file_path"]),
        {"host": "introducer.chiknetwork.com", "port": 9678},
        [],
        0,
        chik_server.config["selected_network"],
        None,
        Logger("node_discovery_tests"),
    )
    assert discovery0 is not None
    assert discovery0.enable_private_networks is False
    await discovery0.initialize_address_manager()
    assert discovery0.address_manager is not None
    assert discovery0.address_manager.allow_private_subnets is False

    # Test with enable_private_networks set to False in Config
    discovery1 = FullNodeDiscovery(
        chik_server,
        0,
        SIMULATOR_ROOT_PATH / Path(chik_server.config["peers_file_path"]),
        {"host": "introducer.chiknetwork.com", "port": 9678, "enable_private_networks": False},
        [],
        0,
        chik_server.config["selected_network"],
        None,
        Logger("node_discovery_tests"),
    )
    assert discovery1 is not None
    assert discovery1.enable_private_networks is False
    await discovery1.initialize_address_manager()
    assert discovery1.address_manager is not None
    assert discovery1.address_manager.allow_private_subnets is False

    # Test with enable_private_networks set to True in Config
    discovery2 = FullNodeDiscovery(
        chik_server,
        0,
        SIMULATOR_ROOT_PATH / Path(chik_server.config["peers_file_path"]),
        {"host": "introducer.chiknetwork.com", "port": 9678, "enable_private_networks": True},
        [],
        0,
        chik_server.config["selected_network"],
        None,
        Logger("node_discovery_tests"),
    )
    assert discovery2 is not None
    assert discovery2.enable_private_networks is True
    await discovery2.initialize_address_manager()
    assert discovery2.address_manager is not None
    assert discovery2.address_manager.allow_private_subnets is True
